import sys
import os
import numpy as np
import h5py
from .utils import parse_params_values
from ..utils import is_writeable
from .cli_configs import PCAFlatsConfig
from .. import version
from ..preproc.flatfield import PCAFlatsDecomposer
from ..io.reader import NXDarksFlats


def get_flats_darks_in_nx(filename):
    dfreader = NXDarksFlats(filename)
    darks = np.concatenate([d for d in dfreader.get_raw_darks()], axis=0)
    flats = np.concatenate([f for f in dfreader.get_raw_flats()], axis=0)
    entry = dfreader.flats_reader.data_path.lstrip("/").split("/")[0]
    return flats, darks, entry


def get_flats_darks_from_h5(filename):
    flats = []
    darks = []
    with h5py.File(filename, "r") as f:
        for k, v in f.items():
            if k == "1.1":
                detector_name = decode_bytes(f["1.1/technique/tomoconfig/detector"][()][0])
            else:
                try:
                    image_key = v["technique/image_key"][()]
                except:
                    raise NotImplementedError(
                        "Legacy h5 file format is not handled. The entry of the h5 file should contain a 'technique/image_key' group."
                    )
                if image_key == 2:  # Darks
                    darks.append(v[f"instrument/{detector_name}/data"][()])
                elif image_key == 1:  # Flats
                    flats.append(v[f"instrument/{detector_name}/data"][()])

    flats = np.concatenate([f for f in flats], axis=0)
    darks = np.concatenate([d for d in darks], axis=0)
    return flats, darks, "entry0000"  # TODO this will be problematic on the reconstruction side


def pcaflats_decomposition(
    flats, darks, pcaflats_filename="PCAFlats.h5", overwrite=False, entry="entry0000", nsigma=3.0
):
    """Compute the PCS decomposition of a series of flats and darks, possibly taken from various scans."""
    try:
        decomposer = PCAFlatsDecomposer(flats, darks, nsigma=nsigma)
        decomposer.save_decomposition(pcaflats_filename, overwrite=overwrite, entry=entry)
        success = True
    except:
        success = False
        raise ValueError("An error occured in the PCA deccomposition.")
    return success


def decode_bytes(content):
    if isinstance(content, bytes):
        return content.decode()
    else:
        return content


def main(argv=None):
    """Compute PCA Flats on a series of datasets (h5 or NX)."""
    if argv is None:
        argv = sys.argv[1:]

    args = parse_params_values(
        PCAFlatsConfig,
        parser_description=f"Compute a PCA Decomposition of flats acquired from various datasets..",
        program_version="nabu " + version,
        user_args=argv,
    )

    # Get "where to write".
    if args["output_filename"] is None:
        abspath = os.path.abspath("./PCAFlats.hdf5")
    else:
        abspath = os.path.abspath(args["output_filename"])

    pcaflats_dir = os.path.dirname(abspath)
    pcaflats_filename = os.path.basename(abspath)

    if is_writeable(pcaflats_dir):
        output_path = os.path.join(pcaflats_dir, pcaflats_filename)
    else:
        raise ValueError(f"Output dir {pcaflats_dir} is not writeable.")

    # raise error if file exists and overwrite=False
    if not args["overwrite"] and os.path.exists(output_path):
        raise FileExistsError(f"Output file {output_path} already exists. Use --overwrite to overwrite it.")

    # Collect raw darks and flats
    flats_stack = []
    darks_stack = []

    for dataset in args["datasets"]:
        filename = os.path.basename(dataset)
        kind = filename.split(".")[-1]
        if kind == "nx":
            flats, darks, entry = get_flats_darks_in_nx(dataset)
        elif kind in ("h5", "hdf5"):
            flats, darks, entry = get_flats_darks_from_h5(dataset)

        flats_stack.append(flats)
        darks_stack.append(darks)

    flats = np.concatenate(flats_stack, axis=0)
    darks = np.concatenate(darks_stack, axis=0)

    exit(
        pcaflats_decomposition(
            flats, darks, pcaflats_filename=output_path, overwrite=args["overwrite"], entry=entry, nsigma=args["nsigma"]
        )
    )


if __name__ == "__main__":
    main()



from .postprocessing import PostProcessingStitchingDumper
from .postprocessing import PostProcessingStitchingDumperNoDD
from .postprocessing import PostProcessingStitchingDumperWithCache
from .preprocessing import PreProcessingStitchingDumper



from dataclasses import dataclass
from itertools import product
import tarfile
import os
import numpy as np
from scipy.signal.windows import gaussian
from silx.resources import ExternalResources
from silx.io.dictdump import nxtodict, dicttonx
from nxtomo.application.nxtomo import ImageKey

utilstest = ExternalResources(
    project="nabu", url_base="http://www.silx.org/pub/nabu/data/", env_key="NABU_DATA", timeout=60
)

__big_testdata_dir__ = os.environ.get("NABU_BIGDATA_DIR")
if __big_testdata_dir__ is None or not (os.path.isdir(__big_testdata_dir__)):
    __big_testdata_dir__ = None

__do_long_tests__ = os.environ.get("NABU_LONG_TESTS", False)  # noqa: PLW1508
if __do_long_tests__:
    try:
        __do_long_tests__ = bool(int(__do_long_tests__))
    except:
        __do_long_tests__ = False

__do_large_mem_tests__ = os.environ.get("NABU_LARGE_MEM_TESTS", False)  # noqa: PLW1508
if __do_large_mem_tests__:
    try:
        __do_large_mem_tests__ = bool(int(__do_large_mem_tests__))
    except:
        __do_large_mem_tests__ = False


def generate_tests_scenarios(configurations):
    """
    Generate "scenarios" of tests.

    The parameter is a dictionary where:
      - the key is the name of a parameter
      - the value is a list of possible parameters

    This function returns a list of dictionary where:
      - the key is the name of a parameter
      - the value is one value of this parameter
    """
    scenarios = [{key: val for key, val in zip(configurations.keys(), p_)} for p_ in product(*configurations.values())]
    return scenarios


def get_data(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    return np.load(dataset_downloaded_path)


@dataclass
class SimpleNXTomoDescription:
    n_darks: int = 0
    n_flats1: int = 0
    n_projs: int = 0
    n_flats2: int = 0
    n_align: int = 0
    frame_shape: tuple = None
    dtype: np.dtype = np.uint16


def get_dummy_nxtomo_info():
    nx_fname = utilstest.getfile("dummy_nxtomo.nx")
    data_desc = SimpleNXTomoDescription(
        n_darks=10, n_flats1=11, n_projs=100, n_flats2=11, n_align=12, frame_shape=(11, 10), dtype=np.uint16
    )
    image_key = np.concatenate(
        [
            np.zeros(data_desc.n_darks, dtype=np.int32) + ImageKey.DARK_FIELD.value,
            np.zeros(data_desc.n_flats1, dtype=np.int32) + ImageKey.FLAT_FIELD.value,
            np.zeros(data_desc.n_projs, dtype=np.int32) + ImageKey.PROJECTION.value,
            np.zeros(data_desc.n_flats2, dtype=np.int32) + ImageKey.FLAT_FIELD.value,
            np.zeros(data_desc.n_align, dtype=np.int32) + ImageKey.ALIGNMENT.value,
        ]
    )
    projs_vals = np.arange(data_desc.n_projs) + data_desc.n_flats1 + data_desc.n_darks
    darks_vals = np.arange(data_desc.n_darks)
    flats1_vals = np.arange(data_desc.n_darks, data_desc.n_darks + data_desc.n_flats1)
    flats2_vals = np.arange(data_desc.n_darks, data_desc.n_darks + data_desc.n_flats2)
    return nx_fname, data_desc, image_key, projs_vals, darks_vals, flats1_vals, flats2_vals


def get_array_of_given_shape(img, shape, dtype):
    """
    From a given image, returns an array of the wanted shape and dtype.
    """

    # Tile image until it's big enough.
    # "fun" fact: using any(blabla) crashes but using any([blabla]) does not, because of variables re-evaluation
    while any([i_dim <= s_dim for i_dim, s_dim in zip(img.shape, shape)]):
        img = np.tile(img, (2, 2))
    if len(shape) == 1:
        arr = img[: shape[0], 0]
    elif len(shape) == 2:
        arr = img[: shape[0], : shape[1]]
    else:
        arr = np.tile(img, (shape[0], 1, 1))[: shape[0], : shape[1], : shape[2]]
    return np.ascontiguousarray(np.squeeze(arr), dtype=dtype)


def get_big_data(filename):
    if __big_testdata_dir__ is None:
        return None
    return np.load(os.path.join(__big_testdata_dir__, filename))


def uncompress_file(compressed_file_path, target_directory):
    if not tarfile.is_tarfile(compressed_file_path):
        raise ValueError(f"Invalid tar file: {compressed_file_path}")

    def is_safe_member(member, target_directory):
        """Ensure the member does not extract outside the target directory."""
        if not isinstance(member, tarfile.TarInfo):
            return False  # Reject any unexpected type

        abs_target = os.path.abspath(target_directory)
        member_path = os.path.abspath(os.path.join(target_directory, member.name))
        return member_path.startswith(abs_target)

    def get_valid_members(tar):
        members = [m for m in tar.getmembers() if is_safe_member(m, target_directory)]
        if not members:
            raise ValueError("No valid files to extract or archive contains unsafe paths.")
        for member in members:
            if not is_safe_member(member, target_directory):
                raise ValueError(f"Unsafe path detected: {member.name}")

    with tarfile.open(compressed_file_path, "r") as tar:
        tar.extractall(  # noqa: S202 - what can be done in addition of the above checks ?
            path=target_directory, members=get_valid_members(tar)
        )


def get_file(fname):
    downloaded_file = utilstest.getfile(fname)
    if ".tar" in fname:
        uncompress_file(downloaded_file, os.path.dirname(downloaded_file))
        downloaded_file = downloaded_file.split(".tar")[0]
    return downloaded_file


def compare_arrays(arr1, arr2, tol, diff=None, absolute_value=True, percent=None, method="max", return_residual=False):
    """
    Utility to compare two arrays.

    Parameters
    ----------
    arr1: numpy.ndarray
        First array to compare
    arr2: numpy.ndarray
        Second array to compare
    tol: float
        Tolerance indicating whether arrays are close to each other.
    diff: numpy.ndarray, optional
        Difference `arr1 - arr2`. If provided, this array is taken instead of `arr1`
        and `arr2`.
    absolute_value: bool, optional
        Whether to take absolute value of the difference.
    percent: float
        If set, a "relative" comparison is performed instead of a subtraction:
        `red(|arr1 - arr2|) / (red(|arr1|) * percent) < tol`
        where "red" is the reduction method (mean, max or median).
    method:
        Reduction method. Can be "max", "mean", or "median".

    Returns
    --------
    (is_close, residual) if return_residual is set to True
    is_close otherwise

    Examples
    --------
    When using method="mean" and absolute_value=True, this function computes
    the Mean Absolute Difference (MAD) metric.
    When also using percent=1.0, this computes the Relative Mean Absolute Difference
    (RMD) metric.
    """
    reductions = {
        "max": np.max,
        "mean": np.mean,
        "median": np.median,
    }
    if method not in reductions:
        raise ValueError("reduction method should be in %s" % str(list(reductions.keys())))
    if diff is None:
        diff = arr1 - arr2
    if absolute_value is not None:
        diff = np.abs(diff)
    residual = reductions[method](diff)
    if percent is not None:
        a1 = np.abs(arr1) if absolute_value else arr1
        residual /= reductions[method](a1)

    res = residual < tol
    if return_residual:
        res = res, residual
    return res


def gaussian_apodization_window(shape, fwhm_ratio=0.7):
    fwhm = fwhm_ratio * np.array(shape)
    sigma = fwhm / 2.355
    return np.outer(*[gaussian(n, s) for n, s in zip(shape, sigma)])


def compare_shifted_images(img1, img2, fwhm_ratio=0.7, return_upper_bound=False):
    """
    Compare two images that are slightly shifted from one another.
    Typically, tomography reconstruction wight slightly different CoR.
    Each image is Fourier-transformed, and the modulus is taken to get rid of the shift between the images.
    An apodization is done to filter the high frequencies that are usually less relevant.

    Parameters
    ----------
    img1: numpy.ndarray
        First image
    img2: numpy.ndarray
        Second image
    fwhm_ratio: float, optional
        Ratio defining the apodization in the frequency domain.
        A small value (eg. 0.2) means that essentually only the low frequencies will be compared.
        A value of 1.0 means no apodization
    return_upper_bound: bool, optional
        Whether to return a (coarse) upper bound of the comparison metric

    Notes
    -----
    This function roughly computes
        |phi(F(img1)) - phi(F(img2))|
    where F is the absolute value of the Fourier transform, and phi some shrinking function (here arcsinh).
    """

    def _fourier_transform(img):
        return np.arcsinh(np.abs(np.fft.fftshift(np.fft.fft2(img))))

    diff = _fourier_transform(img1) - _fourier_transform(img2)
    diff *= gaussian_apodization_window(img1.shape, fwhm_ratio=fwhm_ratio)
    res = np.abs(diff).max()
    if return_upper_bound:
        data_range = np.max(np.abs(img1))
        return res, np.arcsinh(np.prod(img1.shape) * data_range)
    else:
        return res


# To be improved
def generate_nx_dataset(out_fname, image_key, data_volume=None, rotation_angle=None):
    nx_template_file = get_file("dummy.nx.tar.gz")
    nx_dict = nxtodict(nx_template_file)
    nx_entry = nx_dict["entry"]

    def _get_field(dict_, path):
        path = path.strip("/")
        split_path = path.split("/")
        if len(split_path) == 1:
            return dict_[split_path[0]]
        return _get_field(dict_[split_path[0]], "/".join(split_path[1:]))

    for name in ["image_key", "image_key_control"]:
        nx_entry["data"][name] = image_key
        nx_entry["instrument"]["detector"][name] = image_key

    if rotation_angle is not None:
        nx_entry["data"]["rotation_angle"] = rotation_angle
        nx_entry["sample"]["rotation_angle"] = rotation_angle

    dicttonx(nx_dict, out_fname)



import os
from threading import get_ident
from multiprocessing.pool import ThreadPool
from posixpath import sep as posix_sep, join as posix_join
import numpy as np
from h5py import Dataset
from silx.io import get_data
from silx.io.dictdump import h5todict
from tomoscan.io import HDF5File
from .utils import convert_dict_values, get_first_hdf5_entry
from ..misc.binning import binning as image_binning
from ..utils import (
    check_supported,
    deprecated,
    deprecation_warning,
    indices_to_slices,
    compacted_views,
    merge_slices,
    get_3D_subregion,
    get_shape_from_sliced_dims,
    get_size_from_sliced_dimension,
    safe_format,
)

try:
    from fabio.edfimage import EdfImage
except ImportError:
    EdfImage = None


class Reader:
    """
    Abstract class for various file readers.
    """

    def __init__(self, sub_region=None):
        """
        Parameters
        ----------
        sub_region: tuple, optional
            Coordinates in the form (start_x, end_x, start_y, end_y), to read
            a subset of each frame. It can be used for Regions of Interest (ROI).
            Indices start at zero !
        """
        self._set_default_parameters(sub_region)

    def _set_default_parameters(self, sub_region):
        self._set_subregion(sub_region)

    def _set_subregion(self, sub_region):
        self.sub_region = sub_region
        if sub_region is not None:
            start_x, end_x, start_y, end_y = sub_region
            self.start_x = start_x
            self.end_x = end_x
            self.start_y = start_y
            self.end_y = end_y
        else:
            self.start_x = 0
            self.end_x = None
            self.start_y = 0
            self.end_y = None

    def get_data(self, data_url):
        """
        Get data from a silx.io.url.DataUrl
        """
        raise ValueError("Base class")

    def release(self):
        """
        Release the file if needed.
        """
        pass


class NPReader(Reader):
    multi_load = True

    def __init__(self, sub_region=None, mmap=True):
        """
        Reader for NPY/NPZ files. Mostly used for internal development.
        Please refer to the documentation of nabu.io.reader.Reader
        """
        super().__init__(sub_region=sub_region)
        self._file_desc = {}
        self._set_mmap(mmap)

    def _set_mmap(self, mmap):
        self.mmap_mode = "r" if mmap else None

    def _open(self, data_url):
        file_path = data_url.file_path()
        file_ext = self._get_file_type(file_path)
        if file_ext == "npz":
            if file_path not in self._file_desc:
                self._file_desc[file_path] = np.load(file_path, mmap_mode=self.mmap_mode)
            data_ref = self._file_desc[file_path][data_url.data_path()]
        else:
            data_ref = np.load(file_path, mmap_mode=self.mmap_mode)
        return data_ref

    @staticmethod
    def _get_file_type(fname):
        if fname.endswith(".npy"):
            return "npy"
        elif fname.endswith(".npz"):
            return "npz"
        else:
            raise ValueError("Not a numpy file: %s" % fname)

    def get_data(self, data_url):
        data_ref = self._open(data_url)
        data_slice = data_url.data_slice()
        if data_slice is None:
            res = data_ref[self.start_y : self.end_y, self.start_x : self.end_x]
        else:
            res = data_ref[data_slice, self.start_y : self.end_y, self.start_x : self.end_x]
        return res

    def release(self):
        for fname, fdesc in self._file_desc.items():
            if fdesc is not None:
                fdesc.close()
                self._file_desc[fname] = None

    def __del__(self):
        self.release()


class EDFReader(Reader):
    multi_load = False  # not implemented

    def __init__(self, sub_region=None):
        """
        A class for reading series of EDF Files.
        Multi-frames EDF are not supported.
        """
        if EdfImage is None:
            raise ImportError("Need fabio to use this reader")
        super().__init__(sub_region=sub_region)
        self._reader = EdfImage()
        self._first_fname = None

    def read(self, fname):
        if self._first_fname is None:
            self._first_fname = fname
            self._reader.read(fname)
        if self.sub_region is None:
            data = self._reader.data
        else:
            data = self._reader.fast_read_roi(fname, (slice(self.start_y, self.end_y), slice(self.start_x, self.end_x)))
        # self._reader.close()
        return data

    def get_data(self, data_url):
        return self.read(data_url.file_path())


class HDF5Reader(Reader):
    multi_load = True

    def __init__(self, sub_region=None):
        """
        A class for reading a HDF5 File.
        """
        super().__init__(sub_region=sub_region)
        self._file_desc = {}

    def _open(self, file_path):
        if file_path not in self._file_desc:
            self._file_desc[file_path] = HDF5File(file_path, "r", swmr=True)

    def get_data(self, data_url):
        file_path = data_url.file_path()
        self._open(file_path)
        h5dataset = self._file_desc[file_path][data_url.data_path()]
        data_slice = data_url.data_slice()
        if data_slice is None:
            res = h5dataset[self.start_y : self.end_y, self.start_x : self.end_x]
        else:
            res = h5dataset[data_slice, self.start_y : self.end_y, self.start_x : self.end_x]

        return res

    def release(self):
        for fname, fdesc in self._file_desc.items():
            if fdesc is not None:
                try:
                    fdesc.close()
                    self._file_desc[fname] = None
                except Exception as exc:
                    print("Error while closing %s: %s" % (fname, str(exc)))

    def __del__(self):
        self.release()


class HDF5Loader:
    """
    An alternative class to HDF5Reader where information is first passed at class instantiation
    """

    def __init__(self, fname, data_path, sub_region=None, data_buffer=None, pre_allocate=True, dtype="f"):
        self.fname = fname
        self.data_path = data_path
        self._set_subregion(sub_region)
        if not ((data_buffer is not None) ^ (pre_allocate is True)):
            raise ValueError("Please provide either 'data_buffer' or 'pre_allocate'")
        self.data = data_buffer
        self._loaded = False
        self.expected_shape = get_hdf5_dataset_shape(fname, data_path, sub_region=sub_region)
        if pre_allocate:
            self.data = np.zeros(self.expected_shape, dtype=dtype)

    def _set_subregion(self, sub_region):
        self.sub_region = sub_region
        if sub_region is not None:
            start_z, end_z, start_y, end_y, start_x, end_x = sub_region
            self.start_x, self.end_x = start_x, end_x
            self.start_y, self.end_y = start_y, end_y
            self.start_z, self.end_z = start_z, end_z
        else:
            self.start_x, self.end_x = None, None
            self.start_y, self.end_y = None, None
            self.start_z, self.end_z = None, None

    def load_data(self, force_load=False, output=None):
        if self._loaded and not force_load:
            return self.data
        output = self.data if output is None else output
        with HDF5File(self.fname, "r") as fdesc:
            if output is None:
                output = fdesc[self.data_path][
                    self.start_z : self.end_z, self.start_y : self.end_y, self.start_x : self.end_x
                ]
            else:
                output[:] = fdesc[self.data_path][
                    self.start_z : self.end_z, self.start_y : self.end_y, self.start_x : self.end_x
                ]
        self._loaded = True
        return output


class VolReaderBase:
    """
    Base class with common code for data readers (EDFStackReader, NXTomoReader, etc)
    """

    def __init__(self, *args, **kwargs):
        raise ValueError("Base class")

    def _set_subregion(self, sub_region):
        slice_angle, slice_z, slice_x = (None, None, None)
        if isinstance(sub_region, slice):
            # Assume selection is done only along dimension 0
            slice_angle = sub_region
            slice_z = None
            slice_x = None
        if isinstance(sub_region, (tuple, list)):
            slice_angle, slice_z, slice_x = sub_region
        self.sub_region = (
            slice_angle if slice_angle is not None else slice(None, None),
            slice_z if slice_z is not None else slice(None, None),
            slice_x if slice_x is not None else slice(None, None),
        )

    def _set_processing_function(self, processing_func, processing_func_args, processing_func_kwargs):
        self.processing_func = processing_func
        self._processing_func_args = processing_func_args or []
        self._processing_func_kwargs = processing_func_kwargs or {}

    def _get_output(self, array):
        if array is not None:
            if array.shape != self.output_shape:
                raise ValueError("Expected output shape %s but got %s" % (self.output_shape, array.shape))
            if array.dtype != self.output_dtype:
                raise ValueError("Expected output dtype '%s' but got '%s'" % (self.output_dtype, array.dtype))
            output = array
        else:
            output = np.zeros(self.output_shape, dtype=self.output_dtype)
        return output

    def get_frames_indices(self):
        return np.arange(self.data_shape_total[0])[self._source_selection[0]]


class NXTomoReader(VolReaderBase):
    image_key_path = "instrument/detector/image_key_control"
    multiple_frames_per_file = True

    def __init__(
        self,
        filename,
        data_path="{entry}/instrument/detector/data",
        sub_region=None,
        image_key=0,
        output_dtype=np.float32,
        processing_func=None,
        processing_func_args=None,
        processing_func_kwargs=None,
    ):
        """
        Read a HDF5 file in NXTomo layout.

        Parameters
        ----------
        filename: str
            Path to the file to read.
        data_path: str
            Path within the HDF5 file, eg. "entry/instrument/detector/data".
            Default is {entry}/data/data where {entry} is a magic keyword for the first entry.
        sub_region: slice or tuple, optional
            Region to select within the data, once the "image key" selection has been done.
            If None, all the data (corresponding to image_key) is selected.
            If slice(start, stop) is provided, the selection is done along dimension 0.
            Otherwise, it must be a 3-tuple of slices in the form
            (slice(start_angle, end_angle, step_angle), slice(start_z, end_z, step_z), slice(start_x, end_x, step_x))
            Each of the parameters can be None, in this case the default start and end are taken in each dimension.
        output_dtype: numpy.dtype, optional
            Output data type if the data memory is allocated by this class. Default is float32.
        image_key: int, or None, optional
            Image type to read (see NXTomo documentation). 0 for projections, 1 for flat-field, 2 for dark field.
            If set to None, all the data will be read.
        processing_func: callable, optional
            Function to be called on each loaded stack of images.
            If provided, this function first argument must be the source buffer (3D array: stack of raw images),
            and the second argument must be the destination buffer (3D array, stack of output images). It can be None.

        Other Parameters
        ----------------
        The other parameters are passed to "processing_func" if this parameter is not None.

        """
        self.filename = filename
        self.data_path = safe_format(data_path or "", entry=get_first_hdf5_entry(filename))
        self._set_image_key(image_key)
        self._set_subregion(sub_region)
        self._get_shape()
        self._set_processing_function(processing_func, processing_func_args, processing_func_kwargs)
        self._get_source_selection()
        self._init_output(output_dtype)

    def _get_input_dtype(self):
        return get_hdf5_dataset_dtype(self.filename, self.data_path)

    def _init_output(self, output_dtype):
        output_shape = get_shape_from_sliced_dims(self.data_shape_total, self._source_selection)
        self.output_dtype = output_dtype
        self._output_shape_no_processing = output_shape
        if self.processing_func is not None:
            test_subvolume = np.zeros((1,) + output_shape[1:], dtype=self._get_input_dtype())
            out = self.processing_func(
                test_subvolume, None, *self._processing_func_args, **self._processing_func_kwargs
            )
            output_image_shape = out.shape[1:]
            output_shape = (output_shape[0],) + output_image_shape
        self.output_shape = output_shape
        self._tmp_dst_buffer = None

    def _set_image_key(self, image_key):
        self.image_key = image_key
        entry = get_entry_from_h5_path(self.data_path)
        image_key_path = posix_join(entry, self.image_key_path)
        with HDF5File(self.filename, "r") as f:
            image_key_val = f[image_key_path][()]
        idx = np.where(image_key_val == image_key)[0]
        if len(idx) == 0:
            raise FileNotFoundError("No frames found with image key = %d" % image_key)
        self._image_key_slices = indices_to_slices(idx)

    def _get_shape(self):
        # Shape of the total HDF5-NXTomo data (including darks and flats)
        self.data_shape_total = get_hdf5_dataset_shape(self.filename, self.data_path)
        # Shape of the data once the "image key" is selected
        n_imgs = self.data_shape_total[0]
        self.data_shape_imagekey = (
            sum([get_size_from_sliced_dimension(n_imgs, slice_) for slice_ in self._image_key_slices]),
        ) + self.data_shape_total[1:]

        # Shape of the data after sub-regions are selected
        self.data_shape_subregion = get_shape_from_sliced_dims(self.data_shape_imagekey, self.sub_region)

    def _get_source_selection(self):
        if len(self._image_key_slices) == 1 and self.processing_func is None:
            # Simple case:
            #   - One single chunk to load, i.e len(self._image_key_slices) == 1
            #   - No on-the-fly processing (binning, distortion correction, ...)
            # In this case, we can use h5py read_direct() to avoid extraneous memory consumption
            image_key_slice = self._image_key_slices[0]
            # merge image key selection and user selection (if any)
            angles_slice = self.sub_region[0]
            if isinstance(angles_slice, slice) or angles_slice is None:
                angles_slice = merge_slices(image_key_slice, self.sub_region[0] or slice(None, None))
            else:  # assuming numpy array
                # TODO more elegant
                angles_slice = np.arange(self.data_shape_total[0], dtype=np.uint64)[image_key_slice][angles_slice]
            self._source_selection = (angles_slice,) + self.sub_region[1:]
        else:
            user_selection_dim0 = self.sub_region[0]
            indices = np.arange(self.data_shape_total[0])
            data_selection_indices_axis0 = np.hstack(
                [indices[image_key_slice] for image_key_slice in self._image_key_slices]
            )[user_selection_dim0]

            self._source_selection = (data_selection_indices_axis0,) + self.sub_region[1:]

    def _get_temporary_buffer(self, convert_after_reading):
        if self._tmp_dst_buffer is None:
            shape = self._output_shape_no_processing
            dtype = self.output_dtype if not (convert_after_reading) else self._get_input_dtype()
            self._tmp_dst_buffer = np.zeros(shape, dtype=dtype)
        return self._tmp_dst_buffer

    def load_data(self, output=None, convert_after_reading=True):
        """
        Read data.

        Parameters
        -----------
        output: array-like, optional
            Destination 3D array that will hold the data.
            If provided, use this memory buffer instead of allocating the memory.
            Its shape must be compatible with the selection of 'sub_region' and 'image_key'.
        conver_after_reading: bool, optional
            Whether to do the dtype conversion (if any, eg. uint16 to float32) after data reading.
            With using h5py's read_direct(), reading from uint16 to float32 is extremely slow,
            so data type conversion should be done after reading.
            The drawback is that it requires more memory.
        """
        output = self._get_output(output)
        convert_after_reading &= np.dtype(self.output_dtype) != np.dtype(self._get_input_dtype())
        if convert_after_reading or self.processing_func is not None:
            dst_buffer = self._get_temporary_buffer(convert_after_reading)
        else:
            dst_buffer = output

        with HDF5File(self.filename, "r") as f:
            dptr = f[self.data_path]
            dptr.read_direct(
                dst_buffer,
                source_sel=self._source_selection,
                dest_sel=None,
            )
        if self.processing_func is not None:
            self.processing_func(dst_buffer, output, *self._processing_func_args, **self._processing_func_kwargs)
        elif dst_buffer.ctypes.data != output.ctypes.data:
            output[:] = dst_buffer[:]  # cast

        return output


class NXDarksFlats:

    _reduce_func = {
        "median": np.median,
        "mean": np.mean,
    }

    def __init__(self, filename, **nxtomoreader_kwargs):
        nxtomoreader_kwargs.pop("image_key", None)
        self.darks_reader = NXTomoReader(filename, image_key=2, **nxtomoreader_kwargs)
        self.flats_reader = NXTomoReader(filename, image_key=1, **nxtomoreader_kwargs)
        self._raw_darks = None
        self._raw_flats = None

    def _get_raw_frames(self, what, force_reload=False, as_multiple_array=True):
        check_supported(what, ["darks", "flats"], "frame type")
        loaded_frames = getattr(self, "_raw_%s" % what)
        reader = getattr(self, "%s_reader" % what)
        if force_reload or loaded_frames is None:
            loaded_frames = reader.load_data()
            setattr(self, "_raw_%s" % what, loaded_frames)
        res = loaded_frames
        if as_multiple_array:
            slices_ = compacted_views(reader._image_key_slices)
            return [res[slice_] for slice_ in slices_]
        return res

    def _get_reduced_frames(self, what, method="mean", force_reload=False, as_dict=False):
        raw_frames = self._get_raw_frames(what, force_reload=force_reload, as_multiple_array=True)
        reduced_frames = [self._reduce_func[method](frames, axis=0) for frames in raw_frames]
        reader = getattr(self, "%s_reader" % what)
        if as_dict:
            return {k: v for k, v in zip([s.start for s in reader._image_key_slices], reduced_frames)}
        return reduced_frames

    def get_raw_darks(self, force_reload=False, as_multiple_array=True):
        return self._get_raw_frames("darks", force_reload=force_reload, as_multiple_array=as_multiple_array)

    def get_raw_flats(self, force_reload=False, as_multiple_array=True):
        return self._get_raw_frames("flats", force_reload=force_reload, as_multiple_array=as_multiple_array)

    def get_reduced_darks(self, method="mean", force_reload=False, as_dict=False):
        return self._get_reduced_frames("darks", method=method, force_reload=force_reload, as_dict=as_dict)

    def get_reduced_flats(self, method="median", force_reload=False, as_dict=False):
        return self._get_reduced_frames("flats", method=method, force_reload=force_reload, as_dict=as_dict)

    def get_raw_current(self, h5_path="{entry}/control/data"):
        h5_path = safe_format(h5_path, entry=self.flats_reader.data_path.split(posix_sep)[0])
        with HDF5File(self.flats_reader.filename, "r") as f:
            current = f[h5_path][()]
        return {sl.start: current[sl] for sl in self.flats_reader._image_key_slices}

    def get_reduced_current(self, h5_path="{entry}/control/data", method="median"):
        current = self.get_raw_current(h5_path=h5_path)
        return {k: self._reduce_func[method](current[k]) for k in current}


class EDFStackReader(VolReaderBase):
    multiple_frames_per_file = False

    def __init__(
        self,
        filenames,
        sub_region=None,
        output_dtype=np.float32,
        n_reading_threads=1,
        processing_func=None,
        processing_func_args=None,
        processing_func_kwargs=None,
    ):
        self.filenames = filenames
        self.n_reading_threads = n_reading_threads
        self._set_subregion(sub_region)
        self._get_shape()
        self._set_processing_function(processing_func, processing_func_args, processing_func_kwargs)
        self._get_source_selection()
        self._init_output(output_dtype)

    def _get_input_dtype(self):
        return EDFReader().read(self.filenames[0]).dtype

    def _get_shape(self):
        first_filename = self.filenames[0]

        # Shape of the total data (without subregion selection)
        reader_all = EDFReader()
        first_frame_full = reader_all.read(first_filename)
        self.data_shape_total = (len(self.filenames),) + first_frame_full.shape
        self.input_dtype = first_frame_full.dtype

        self.data_shape_subregion = get_shape_from_sliced_dims(
            self.data_shape_total, self.sub_region
        )  # might fail if sub_region is not a slice ?

    def _init_output(self, output_dtype):
        output_shape = get_shape_from_sliced_dims(self.data_shape_total, self._source_selection)
        self.output_dtype = output_dtype
        if self.processing_func is not None:
            test_image = np.zeros(output_shape[1:], dtype=self._get_input_dtype())
            out = self.processing_func(test_image, *self._processing_func_args, **self._processing_func_kwargs)
            output_image_shape = out.shape
            output_shape = (output_shape[0],) + output_image_shape
        self.output_shape = output_shape

    def _get_source_selection(self):
        self._source_selection = self.sub_region

        self._sub_region_xy_for_edf_reader = (
            self.sub_region[2].start or 0,
            self.sub_region[2].stop or self.data_shape_total[2],
            self.sub_region[1].start or 0,
            self.sub_region[1].stop or self.data_shape_total[1],
        )
        self.filenames_subsampled = self.filenames[self.sub_region[0]]

    def load_data(self, output=None):
        output = self._get_output(output)

        readers = {}

        def _init_reader_thread():
            readers[get_ident()] = EDFReader(self._sub_region_xy_for_edf_reader)

        def _get_data(i_fname):
            i, fname = i_fname
            reader = readers[get_ident()]
            frame = reader.read(fname)
            if self.processing_func is not None:
                frame = self.processing_func(frame, *self._processing_func_args, **self._processing_func_kwargs)
            output[i] = frame

        with ThreadPool(self.n_reading_threads, initializer=_init_reader_thread) as tp:
            tp.map(
                _get_data,
                zip(
                    range(len(self.filenames_subsampled)),
                    self.filenames_subsampled,
                ),
            )

        return output


Readers = {
    "edf": EDFReader,
    "hdf5": HDF5Reader,
    "h5": HDF5Reader,
    "nx": HDF5Reader,
    "npz": NPReader,
    "npy": NPReader,
}


@deprecated(
    "Function load_images_from_dataurl_dict is deprecated and will be removed in a fugure version", do_print=True
)
def load_images_from_dataurl_dict(data_url_dict, sub_region=None, dtype="f", binning=None):
    """
    Load a dictionary of dataurl into numpy arrays.

    Parameters
    ----------
    data_url_dict: dict
        A dictionary where the keys are integers (the index of each image in the dataset),
        and the values are numpy.ndarray (data_url_dict).
    sub_region: tuple, optional
        Tuple in the form (y_subregion, x_subregion)
        where xy_subregion is a tuple in the form slice(start, stop, step)

    Returns
    --------
    res: dict
        A dictionary where the keys are the same as `data_url_dict`, and the values are numpy arrays.

    Notes
    -----
    This function is used to load flats/darks. Usually, these are reduced flats/darks, meaning that
    'data_url_dict' is a collection of a handful of files (less than 10).
    To load more frames, it would be best to use NXTomoReader / EDFStackReader.
    """
    res = {}
    if sub_region is not None and not isinstance(sub_region[0], slice):
        if len(sub_region) == 4:  # (start_y, end_y, start_x, end_x)
            deprecation_warning(
                "The parameter 'sub_region' was passed as (start_x, end_x, start_y, end_y). This is deprecated and will yield an error in the future. Please use the syntax ((start_z, end_z), (start_x, end_x))",
                do_print=True,
                func_name="load_images_from_dataurl_dict",
            )
            sub_region = (slice(sub_region[2], sub_region[3]), slice(sub_region[0], sub_region[1]))
        else:  # ((start_z, end_z), (start_x, end_x))
            sub_region = tuple(slice(s[0], s[1]) for s in sub_region)
    for idx, data_url in data_url_dict.items():
        frame = get_data(data_url)
        if sub_region is not None:
            frame = frame[sub_region[0], sub_region[1]]
        if binning is not None:
            frame = image_binning(frame, binning, out_dtype=dtype)
        res[idx] = frame
    return res


def load_images_stack_from_hdf5(fname, h5_data_path, sub_region=None):
    """
    Load a 3D dataset from a HDF5 file.

    Parameters
    -----------
    fname: str
        File path
    h5_data_path: str
        Data path within the HDF5 file
    sub_region: tuple, optional
        Tuple indicating which sub-volume to load, in the form (xmin, xmax, ymin, ymax, zmin, zmax)
        where the 3D dataset has the python shape (N_Z, N_Y, N_X).
        This means that the data will be loaded as `data[zmin:zmax, ymin:ymax, xmin:xmax]`.
    """
    xmin, xmax, ymin, ymax, zmin, zmax = get_3D_subregion(sub_region)
    with HDF5File(fname, "r") as f:
        d_ptr = f[h5_data_path]
        data = d_ptr[zmin:zmax, ymin:ymax, xmin:xmax]
    return data


def get_hdf5_dataset_shape(fname, h5_data_path, sub_region=None):
    zmin, zmax, ymin, ymax, xmin, xmax = get_3D_subregion(sub_region)
    with HDF5File(fname, "r") as f:
        d_ptr = f[h5_data_path]
        shape = d_ptr.shape
    n_z, n_y, n_x = shape
    # perhaps there is more elegant
    res_shape = []
    for n, bounds in zip([n_z, n_y, n_x], ((zmin, zmax), (ymin, ymax), (xmin, xmax))):
        res_shape.append(np.arange(n)[bounds[0] : bounds[1]].size)
    return tuple(res_shape)


def get_hdf5_dataset_dtype(fname, h5_data_path):
    with HDF5File(fname, "r") as f:
        d_ptr = f[h5_data_path]
        dtype = d_ptr.dtype
    return dtype


def get_entry_from_h5_path(h5_path):
    v = h5_path.split(posix_sep)
    return v[0] or v[1]


def list_hdf5_entries(fname):
    with HDF5File(fname, "r") as f:
        entries = list(f.keys())
    return entries


def check_virtual_sources_exist(fname, data_path):
    with HDF5File(fname, "r") as f:
        if data_path not in f:
            print("No dataset %s in file %s" % (data_path, fname))
            return False
        dptr = f[data_path]
        if not dptr.is_virtual:
            return True
        for vsource in dptr.virtual_sources():
            vsource_fname = os.path.join(os.path.dirname(dptr.file.filename), vsource.file_name)
            if not os.path.isfile(vsource_fname):
                print("No such file: %s" % vsource_fname)
                return False
            elif not check_virtual_sources_exist(vsource_fname, vsource.dset_name):
                print("Error with virtual source %s" % vsource_fname)
                return False
    return True


def get_hdf5_file_all_virtual_sources(file_path_or_obj, return_only_filenames=False):
    result = []

    def collect_vsources(name, obj):
        if isinstance(obj, Dataset) and obj.is_virtual:
            vs = obj.virtual_sources()
            if return_only_filenames:
                vs = [vs_.file_name for vs_ in vs]
            result.append({name: vs})

    _self_opened_file = False
    if isinstance(file_path_or_obj, str):
        fdesc = HDF5File(file_path_or_obj, "r")
        _self_opened_file = True
    else:
        fdesc = file_path_or_obj

    fdesc.visititems(collect_vsources)

    if _self_opened_file:
        fdesc.close()
    return result


def import_h5_to_dict(h5file, h5path, asarray=False):
    """
    Wrapper on top of silx.io.dictdump.dicttoh5 replacing "None" with None

    Parameters
    -----------
    h5file: str
        File name
    h5path: str
        Path in the HDF5 file
    asarray: bool, optional
        Whether to convert each numeric value to an 0D array. Default is False.
    """
    dic = h5todict(h5file, path=h5path, asarray=asarray)
    modified_dic = convert_dict_values(dic, {"None": None}, bytes_tostring=True)
    return modified_dic



import numpy as np
import pytest
from nabu.testutils import get_data
from nabu.cuda.utils import __has_cupy__

if __has_cupy__:
    from nabu.reconstruction.projection import Projector
try:
    import astra

    __has_astra__ = True
except ImportError:
    __has_astra__ = False


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.image = get_data("brain_phantom.npz")["data"]
    cls.sino_ref = get_data("mri_sino500.npz")["data"]
    cls.n_angles, cls.dwidth = cls.sino_ref.shape
    cls.rtol = 1e-3
    if __has_cupy__:
        ...


@pytest.mark.skipif(not (__has_cupy__), reason="Need cupy for this test")
@pytest.mark.usefixtures("bootstrap")
class TestProjection:
    def check_result(self, img1, img2, err_msg):
        max_diff = np.max(np.abs(img1 - img2))
        assert max_diff / img1.max() < self.rtol, err_msg + " : max diff = %.3e" % max_diff

    def test_proj_simple(self):
        P = Projector(self.image.shape, self.n_angles)
        res = P(self.image)
        self.check_result(res, self.sino_ref, "Something wrong with simple projection")

    def test_input_output_kinds(self):
        P = Projector(self.image.shape, self.n_angles)

        # input on GPU, output on CPU
        d_img = P.cuda_processing.to_device("d_img", self.image)
        res = P(d_img)
        self.check_result(res, self.sino_ref, "Something wrong: input GPU, output CPU")

        # input on CPU, output on GPU
        out = P.cuda_processing.allocate_array("out", P.sino_shape, dtype="f")
        res = P(self.image, output=out)
        self.check_result(out.get(), self.sino_ref, "Something wrong: input CPU, output GPU")

        # input and output on GPU
        out.fill(0)
        P(d_img, output=out)
        self.check_result(out.get(), self.sino_ref, "Something wrong: input GPU, output GPU")

    def test_odd_size(self):
        image = self.image[:511, :]
        P = Projector(image.shape, self.n_angles - 1)
        res = P(image)  # noqa: F841
        # TODO check

    @pytest.mark.skipif(not (__has_astra__), reason="Need astra-toolbox for this test")
    def test_against_astra(self):
        def proj_astra(img, angles, rot_center=None):
            vol_geom = astra.create_vol_geom(img.shape)
            if np.isscalar(angles):
                angles = np.linspace(0, np.pi, angles, False)
            proj_geom = astra.create_proj_geom("parallel", 1.0, img.shape[-1], angles)
            if rot_center is not None:
                cor_shift = (img.shape[-1] - 1) / 2.0 - rot_center
                proj_geom = astra.geom_postalignment(proj_geom, cor_shift)

            projector_id = astra.create_projector("cuda", proj_geom, vol_geom)
            sinogram_id, sinogram = astra.create_sino(img, projector_id)

            astra.data2d.delete(sinogram_id)
            astra.projector.delete(projector_id)
            return sinogram

        # Center of rotation to test
        cors = [None, 255.5, 256, 260, 270.2, 300, 150]

        for cor in cors:
            res_astra = proj_astra(self.image, 500, rot_center=cor)
            res_nabu = Projector(self.image.shape, 500, rot_center=cor).projection(self.image)
            self.check_result(res_nabu, res_astra, "Projection with CoR = %s" % str(cor))



from os import linesep
import numpy as np
from cupy import RawModule
from ..processing.kernel_base import KernelBase
from ..utils import catch_warnings  # TODO use warnings.catch_warnings once python < 3.11 is dropped


class CudaKernel(KernelBase):
    """
    Helper class that wraps CUDA kernel through cupy SourceModule.

    Parameters
    -----------
    kernel_name: str
        Name of the CUDA kernel.
    filename: str, optional
        Path to the file name containing kernels definitions
    src: str, optional
        Source code of kernels definitions
    automation_params: dict, optional
        Automation parameters, see below
    sourcemodule_kwargs: optional
        Extra arguments to provide to cupy.RawModule(),

    """

    def __init__(
        self,
        kernel_name,
        filename=None,
        src=None,
        automation_params=None,
        silent_compilation_warnings=False,
        extern_c=True,
        **sourcemodule_kwargs,
    ):
        super().__init__(
            kernel_name,
            filename=filename,
            src=src,
            automation_params=automation_params,
            silent_compilation_warnings=silent_compilation_warnings,
        )

        if extern_c:
            # pycuda/pyopencl do that automatically, not cupy
            self.src = patch_sourcecode_add_externC(self.src, filename=filename)
        self.compile_kernel_source(kernel_name, sourcemodule_kwargs)

    def compile_kernel_source(self, kernel_name, sourcemodule_kwargs):
        self.sourcemodule_kwargs = sourcemodule_kwargs
        # Use NVCC by default
        if self.sourcemodule_kwargs.get("backend", None) is None:
            self.sourcemodule_kwargs["backend"] = "nvcc"
        #
        self.kernel_name = kernel_name
        with catch_warnings(action=("ignore" if self.silent_compilation_warnings else None)):  # pylint: disable=E1123
            self.module = RawModule(code=self.src, **self.sourcemodule_kwargs)
            self.module.compile()
        self.func = self.module.get_function(kernel_name)

    def follow_device_arr(self, args):
        return args

    def call(self, *args, **kwargs):
        grid, block, args, kwargs = self._prepare_call(*args, **kwargs)
        self.func(grid, block, args)

    __call__ = call


def patch_sourcecode_add_externC(src_code, filename=None):
    """
    Patch a source code to surround the relevant parts with 'extern C {}' directive
    The NVCC compiler needs this to avoid name mangling.
    """
    lines = src_code.split(linesep)
    incl_idx = []
    incl_idx_nows = []

    i = 0
    for i0, line in enumerate(lines):
        line = line.strip()
        if line.startswith(("//", "/*")):
            continue
        if line.startswith("#include"):
            incl_idx.append(i0)
            incl_idx_nows.append(i)
            i += 1
    if len(incl_idx) == 0:
        insertion_idx = 0
    else:
        if np.any(np.diff(incl_idx_nows) > 1):
            raise ValueError(
                f"'#include' should be grouped on top of the file, not separated by anything else - found #include at lines {incl_idx} in file {filename}"
            )
        else:
            insertion_idx = incl_idx[-1] + 1
    lines.insert(insertion_idx, 'extern "C" {')
    lines.append("}")
    modified_src = linesep.join(lines)
    return modified_src



import os
import multiprocessing as mp
import queue
from threading import Thread
from traceback import print_exc
from time import sleep

"""
A simple yet quite generic tasks distribution system using multiprocessing.
There are lots of task distribution frameworks (dask_distributed, ray, etc),
but here the needs is really simple: distribute independent tasks on a few processes, without inter-process communication.
"""


# Possible improvements
#   future.error : bool indicating whether something went wrong
#   __repr__ / __str__  to inspect a future object at glance


class Future:
    """
    A dummy future-like object that mimics concurrent.futures.
    We don't use concurrent.futures because it is tightly coupled to its own executors (ThreadPoolExecutor and ProcessPoolExecutor).
    A concurrent.futures.Future is not a free-standing object that can be safely complete yourself.
    It's designed to be managed by an executor which handles the worker lifecycle, task submission, and result propagation.
    Executors use their own internal queues and bookkeeping; they won't integrate cleanly in this case because
    we use a multiprocessing.Queue-based protocol and persistent per-GPU processes.
    """

    def __init__(self):
        self._result = None
        self._done = mp.Event()
        self._callbacks = []

    def set_result(self, value):
        self._result = value
        self._done.set()
        for cb in self._callbacks:
            cb(self)

    def result(self, timeout=None):
        self._done.wait(timeout)
        return self._result

    def done(self):
        return self._done.is_set()

    def add_done_callback(self, fn):
        if self.done():
            fn(self)
        else:
            self._callbacks.append(fn)


def worker_entrypoint(
    tasks_queue, results_queue, processing_class, worker_id=None, proc_class_init_args=None, proc_class_init_kwargs=None
):
    """
    Persistent CPU worker.

    Work is submitted to this worker by submitting a non-empty tuple (task_id, task) through the 'tasks_queue' Queue.
    A tuple (X, None) will stop the current worker from accepting new tasks.

    Parameters
    ----------
    worker_id: int
        Worker identifier
    tasks_queue: multiprocessing.Queue
        Queue where tasks are sent from main task processor to workers
    results_queue: multiprocessing.Queue
        Queue where results are sent from workers to main task processor
    processing_function: callable
        Function executed by worker each time a task is sent.
        It must have only one argument: the task itself.
    """
    proc = processing_class(*(proc_class_init_args or ()), **(proc_class_init_kwargs or {}))
    proc.worker_id = worker_id
    while True:
        task_id, task = tasks_queue.get()
        if task is None:
            break
        try:
            result = proc.process_task(task)
        except Exception as exc:
            result = {"error": exc}
            print_exc()
        results_queue.put((task_id, result))


# Possible improvements:
#  - submit() could automatically pick a worker
#  - standardize the tasks: dictionary with worker_id
#  - check that 'processing_class' has a 'process_task' method. Make it inherit from a base class ?


class TaskProcessor:

    queue_class = mp.Queue

    def __init__(self, processing_class, n_processes=2, worker_init_args=None, worker_init_kwargs=None):
        self.n_processes = n_processes
        self.task_queues = []
        self.results_queue = self.queue_class()
        self.workers = []
        self.futures = {}
        self._task_counter = 0
        self._workers_stopped = False
        self.processing_class = processing_class
        self._worker_init_args = worker_init_args or ()
        self._worker_init_kwargs = worker_init_kwargs or {}

    def start_workers(self):
        for wid in range(self.n_processes):
            tasks_queue = self.queue_class()
            p = mp.Process(
                target=worker_entrypoint,
                args=(tasks_queue, self.results_queue, self.processing_class),
                kwargs={
                    "worker_id": wid,
                    "proc_class_init_args": self._worker_init_args,
                    "proc_class_init_kwargs": self._worker_init_kwargs,
                },
            )
            p.start()
            self.task_queues.append(tasks_queue)
            self.workers.append(p)

        # start listener thread
        self.listening_thread = Thread(target=self._result_listener, daemon=True).start()

    def start_work(self): ...

    def stop_workers(self):
        if self._workers_stopped:
            return
        for q in self.task_queues:
            q.put((None, None))
        for p in self.workers:
            p.join()
        self._workers_stopped = True

    def _result_listener(self):
        while True:
            task_id, result = self.results_queue.get()
            if task_id not in self.futures:
                break
            self.futures[task_id].set_result(result)

    def submit(self, worker_id, task):
        fut = Future()
        task_id = self._task_counter
        self._task_counter += 1
        self.futures[task_id] = fut
        self.task_queues[worker_id].put((task_id, task))
        return fut


class FakeTaskProcessor(TaskProcessor):
    """
    A "fake" TaskProcessor that does not use multi-processing and will process tasks serially
    """

    queue_class = queue.Queue

    def start_workers(self):
        self.task_queues = [self.queue_class()]  # one single task queue for all workers
        self.workers = [None]

    def stop_workers(self):
        self._workers_stopped = True

    def start_work(self):
        task_queue = self.task_queues[0]
        proc = self.processing_class(*(self._worker_init_args or ()), **(self._worker_init_kwargs or {}))
        proc.worker_id = 0  # ?
        while task_queue.qsize() > 0:
            task_id, task = task_queue.get_nowait()
            if task is None:
                break
            try:
                result = proc.process_task(task)
            except Exception as exc:
                result = {"error": exc}
                print_exc()
            if task_id not in self.futures:
                break
            self.futures[task_id].set_result(result)



# ruff: noqa
try:
    import astra

    __have_astra__ = True
except ImportError:
    __have_astra__ = False
    astra = None


class AstraReconstructor:
    """
    Base class for reconstructors based on the Astra toolbox
    """

    default_extra_options = {
        "axis_correction": None,
        "clip_outer_circle": False,
        "scale_factor": None,
        "filter_cutoff": 1.0,
        "outer_circle_value": 0.0,
    }

    def __init__(
        self,
        sinos_shape,
        angles=None,
        volume_shape=None,
        rot_center=None,
        pixel_size=None,
        padding_mode="zeros",
        filter_name=None,
        slice_roi=None,
        cuda_options=None,
        extra_options=None,
    ):
        self._configure_extra_options(extra_options)
        self._init_cuda(cuda_options)
        self._set_sino_shape(sinos_shape)
        self._orig_prog_geom = None
        self._init_geometry(
            source_origin_dist,
            origin_detector_dist,
            pixel_size,
            angles,
            volume_shape,
            rot_center,
            relative_z_position,
            slice_roi,
        )
        self._init_fdk(padding_mode, filter_name)
        self._alg_id = None
        self._vol_id = None
        self._proj_id = None

    def _configure_extra_options(self, extra_options):
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def _init_cuda(self, cuda_options):
        cuda_options = cuda_options or {}
        self.cuda = CudaProcessing(**cuda_options)

    def _set_sino_shape(self, sinos_shape):
        if len(sinos_shape) != 3:
            raise ValueError("Expected a 3D shape")
        self.sinos_shape = sinos_shape
        self.n_sinos, self.n_angles, self.prj_width = sinos_shape

    def _set_pixel_size(self, pixel_size):
        if pixel_size is None:
            det_spacing_y = det_spacing_x = 1
        elif np.iterable(pixel_size):
            det_spacing_y, det_spacing_x = pixel_size
        else:
            # assuming scalar
            det_spacing_y = det_spacing_x = pixel_size
        self._det_spacing_y = det_spacing_y
        self._det_spacing_x = det_spacing_x

    def _set_slice_roi(self, slice_roi):
        self.slice_roi = slice_roi
        self._vol_geom_n_x = self.n_x
        self._vol_geom_n_y = self.n_y
        self._crop_data = True
        if slice_roi is None:
            return
        start_x, end_x, start_y, end_y = slice_roi
        if roi_is_centered(self.volume_shape[1:], (slice(start_y, end_y), slice(start_x, end_x))):
            # Astra can only reconstruct subregion centered around the origin
            self._vol_geom_n_x = self.n_x - start_x * 2
            self._vol_geom_n_y = self.n_y - start_y * 2
        else:
            raise NotImplementedError(
                "Astra supports only slice_roi centered around origin (got slice_roi=%s with n_x=%d, n_y=%d)"
                % (str(slice_roi), self.n_x, self.n_y)
            )

    def _init_geometry(
        self,
        source_origin_dist,
        origin_detector_dist,
        pixel_size,
        angles,
        volume_shape,
        rot_center,
        relative_z_position,
        slice_roi,
    ):
        if angles is None:
            self.angles = np.linspace(0, 2 * np.pi, self.n_angles, endpoint=True)
        else:
            self.angles = angles
        if volume_shape is None:
            volume_shape = (self.sinos_shape[0], self.sinos_shape[2], self.sinos_shape[2])
        self.volume_shape = volume_shape
        self.n_z, self.n_y, self.n_x = self.volume_shape
        self.source_origin_dist = source_origin_dist
        self.origin_detector_dist = origin_detector_dist
        self.magnification = 1 + origin_detector_dist / source_origin_dist
        self._set_slice_roi(slice_roi)
        self.vol_geom = astra.create_vol_geom(self._vol_geom_n_y, self._vol_geom_n_x, self.n_z)
        self.vol_shape = astra.geom_size(self.vol_geom)
        self._cor_shift = 0.0
        self.rot_center = rot_center
        if rot_center is not None:
            self._cor_shift = (self.sinos_shape[-1] - 1) / 2.0 - rot_center
        self._set_pixel_size(pixel_size)
        self._axis_corrections = self.extra_options.get("axis_correction", None)
        self._create_astra_proj_geometry(relative_z_position)

    def _create_astra_proj_geometry(self, relative_z_position):
        # This object has to be re-created each time, because once the modifications below are done,
        # it is no more a "cone" geometry but a "cone_vec" geometry, and cannot be updated subsequently
        # (see astra/functions.py:271)
        self.proj_geom = astra.create_proj_geom(
            "cone",
            self._det_spacing_x,
            self._det_spacing_y,
            self.n_sinos,
            self.prj_width,
            self.angles,
            self.source_origin_dist,
            self.origin_detector_dist,
        )
        self.relative_z_position = relative_z_position or 0.0
        # This will turn the geometry of type "cone" into a geometry of type "cone_vec"
        if self._orig_prog_geom is None:
            self._orig_prog_geom = self.proj_geom
        self.proj_geom = astra.geom_postalignment(self.proj_geom, (self._cor_shift, 0))
        # (src, detector_center, u, v) = (srcX, srcY, srcZ, dX, dY, dZ, uX, uY, uZ, vX, vY, vZ)
        vecs = self.proj_geom["Vectors"]

        # To adapt the center of rotation:
        # dX = cor_shift * cos(theta) - origin_detector_dist * sin(theta)
        # dY = origin_detector_dist * cos(theta) + cor_shift * sin(theta)
        if self._axis_corrections is not None:
            # should we check that dX and dY match the above formulas ?
            cor_shifts = self._cor_shift + self._axis_corrections
            vecs[:, 3] = cor_shifts * np.cos(self.angles) - self.origin_detector_dist * np.sin(self.angles)
            vecs[:, 4] = self.origin_detector_dist * np.cos(self.angles) + cor_shifts * np.sin(self.angles)

        # To adapt the z position:
        # Component 2 of vecs is the z coordinate of the source, component 5 is the z component of the detector position
        # We need to re-create the same inclination of the cone beam, thus we need to keep the inclination of the two z positions.
        # The detector is centered on the rotation axis, thus moving it up or down, just moves it out of the reconstruction volume.
        # We can bring back the detector in the correct volume position, by applying a rigid translation of both the detector and the source.
        # The translation is exactly the amount that brought the detector up or down, but in the opposite direction.
        vecs[:, 2] = -self.relative_z_position

    def _set_output(self, volume):
        if volume is not None:
            expected_shape = self.vol_shape  # if not (self._crop_data) else self._output_cropped_shape
            self.cuda.check_array(volume, expected_shape)
            self.cuda.set_array("output", volume)
        if volume is None:
            self.cuda.allocate_array("output", self.vol_shape)
        d_volume = self.cuda.get_array("output")
        z, y, x = d_volume.shape
        self._vol_link = astra.data3d.GPULink(d_volume.ptr, x, y, z, d_volume.strides[-2])
        self._vol_id = astra.data3d.link("-vol", self.vol_geom, self._vol_link)

    def _set_input(self, sinos):
        self.cuda.check_array(sinos, self.sinos_shape)
        self.cuda.set_array("sinos", sinos)  # self.cuda.sinos is now a GPU array
        # TODO don't create new link/proj_id if ptr is the same ?
        # But it seems Astra modifies the input sinogram while doing FDK, so this might be not relevant
        d_sinos = self.cuda.get_array("sinos")

        # self._proj_data_link = astra.data3d.GPULink(d_sinos.ptr, self.prj_width, self.n_angles, self.n_z, sinos.strides[-2])
        self._proj_data_link = astra.data3d.GPULink(
            d_sinos.ptr, self.prj_width, self.n_angles, self.n_sinos, d_sinos.strides[-2]
        )
        self._proj_id = astra.data3d.link("-sino", self.proj_geom, self._proj_data_link)

    def _preprocess_data(self):
        d_sinos = self.cuda.sinos
        for i in range(d_sinos.shape[0]):
            self.sino_filter.filter_sino(d_sinos[i], output=d_sinos[i])

    def _update_reconstruction(self):
        cfg = astra.astra_dict("BP3D_CUDA")
        cfg["ReconstructionDataId"] = self._vol_id
        cfg["ProjectionDataId"] = self._proj_id
        if self._alg_id is not None:
            astra.algorithm.delete(self._alg_id)
        self._alg_id = astra.algorithm.create(cfg)

    def reconstruct(self, sinos, output=None, relative_z_position=None):
        """
        sinos: numpy.ndarray or pycuda.gpuarray
            Sinograms, with shape (n_sinograms, n_angles, width)
        output: pycuda.gpuarray, optional
            Output array. If not provided, a new numpy array is returned
        relative_z_position: int, optional
            Position of the central slice of the slab, with respect to the full stack of slices.
            By default it is set to zero, meaning that the current slab is assumed in the middle of the stack
        """
        self._create_astra_proj_geometry(relative_z_position)
        self._set_input(sinos)
        self._set_output(output)
        self._preprocess_data()
        self._update_reconstruction()
        astra.algorithm.run(self._alg_id)
        #
        # NB: Could also be done with
        # from astra.experimental import direct_BP3D
        # projector_id = astra.create_projector("cuda3d", self.proj_geom, self.vol_geom, options=None)
        # direct_BP3D(projector_id, self._vol_link, self._proj_data_link)
        #
        result = self.cuda.get_array("output")
        if output is None:
            result = result.get()
        if self.extra_options.get("scale_factor", None) is not None:
            result *= np.float32(self.extra_options["scale_factor"])  # in-place for pycuda
        self.cuda.recover_arrays_references(["sinos", "output"])
        return result

    def __del__(self):
        if getattr(self, "_alg_id", None) is not None:
            astra.algorithm.delete(self._alg_id)
        if getattr(self, "_vol_id", None) is not None:
            astra.data3d.delete(self._vol_id)
        if getattr(self, "_proj_id", None) is not None:
            astra.data3d.delete(self._proj_id)



from os import remove
import numpy as np
from .. import version
from .reconstruct import get_reconstructor
from ..pipeline.fullfield.reconstruction import NannyPipeline
from .cli_configs import MultiCorConfig
from .utils import parse_params_values
from ..utils import view_as_images_stack


def get_user_cors(cors):
    """
    From a user-provided str describing the centers of rotation, build a list.
    """
    cors = cors.strip("[()]")
    cors = cors.split(",")
    cors = [c.strip() for c in cors]
    cors_list = []
    for c in cors:
        if ":" in c:
            if c.count(":") != 2:
                raise ValueError("Malformed range format for '%s': expected format start:stop:step" % c)
            start, stop, step = c.split(":")
            c_list = np.arange(float(start), float(stop), float(step)).tolist()
        else:
            c_list = [float(c)]
        cors_list.extend(c_list)
    return cors_list


def main():
    args = parse_params_values(
        MultiCorConfig,
        parser_description=f"Perform a tomographic reconstruction of a single slice using multiple centers of rotation",
        program_version="nabu " + version,
    )

    cors = get_user_cors(args["cor"])
    reconstructor = get_reconstructor(
        args,
        # Put a dummy CoR to avoid crash in both full-FoV and extended-FoV.
        # It will be overwritten later by the user-defined CoRs
        overwrite_options={"reconstruction/rotation_axis_position": cors[0]},
    )

    if reconstructor.delta_z > 1:
        raise ValueError("Only slice reconstruction can be used (have delta_z = %d)" % reconstructor.delta_z)

    pipeline_kwargs = {}
    if reconstructor.backend == "cuda":
        pipeline_kwargs["cuda_options"] = reconstructor.cuda_options
    pipeline_kwargs["use_grouped_mode"] = reconstructor._pipeline_mode == "grouped"

    nap = NannyPipeline(
        reconstructor.process_config,
        backend=reconstructor.backend,
        pipeline_mode=reconstructor._pipeline_mode,
        logging_options={"level": reconstructor.logger.level, "logfile": reconstructor.logger.logfile},
        **pipeline_kwargs,
    )
    nap._instantiate_pipeline(reconstructor.tasks[0])
    pipeline = nap.pipeline
    pipeline.process_chunk(reconstructor.tasks[0]["sub_region"])  # warm-up

    file_prefix = pipeline.processing_options["save"]["file_prefix"]
    #####
    # Remove the first reconstructed file (not used here)
    last_file = list(pipeline.writer.writer.browse_data_files())[-1]
    # ruff: noqa: SIM105, S110
    try:
        remove(last_file)
    except:
        pass
    ######

    options = reconstructor.process_config.processing_options["reconstruction"]
    reconstruct_from_sinos_stack = (options["method"].lower() == "cone") or (
        options["method"].lower() == "mlem" and options["implementation"].lower() == "corrct"
    )
    rec_kwargs = {}
    if options["method"].lower() == "cone":
        z_min, z_max = pipeline.sub_region_xz[2:]
        n_z_tot = pipeline.process_config.radio_shape(binning=True)[0]
        z_pos = ((z_min + z_max) / reconstructor.process_config.binning_z / 2) - n_z_tot / 2
        rec_kwargs["relative_z_position"] = z_pos
    do_halftomo = pipeline.process_config.do_halftomo

    rec_instance = pipeline.reconstruction
    # FIXME
    # ConebeamReconstructor will modify in-place the input sinogram when doing FDK, even if non-contiguous (radios layout).
    # FDK uses the 'mult_factor' below that is corrected by SinoFilter.
    # In our case this compensation has to be done once, since we re-use always the same sinogram
    if rec_instance.__class__.__name__ == "ConebeamReconstructor":
        mult_factor = rec_instance.n_angles / 3.141592 * 2
        rec_instance.sino_filter.set_filter(rec_instance.sino_filter.filter_f * mult_factor, normalize=False)
    # ---

    # Get sinogram
    if reconstruct_from_sinos_stack:
        sino = pipeline._d_radios.transpose((1, 0, 2))
    else:
        # Get sinogram into contiguous array
        # TODO Can't do memcpy2D ?! It used to work in cuda 11.
        # For now: transfer to host... not optimal
        sino = pipeline._d_radios[:, pipeline._d_radios.shape[1] // 2, :].get()  # pylint: disable=E1136

    for cor in cors:
        # Re-configure with new CoR
        pipeline.processing_options["reconstruction"]["rotation_axis_position"] = cor
        pipeline.processing_options["save"]["file_prefix"] = file_prefix + "_%.03f" % cor
        pipeline._init_writer(create_subfolder=False, single_output_file_initialized=False)
        # Reconfigure center of rotation
        if not (do_halftomo):
            pipeline.reconstruction.reset_rot_center(cor)
        else:
            # re-initialize FBP object, because in half-tomography the output slice size is a function of CoR
            rec_instance = pipeline.FBPClass(
                sino.shape,
                angles=options["angles"],
                rot_center=cor,
                filter_name=options["fbp_filter_type"] or "none",
                halftomo=options["enable_halftomo"],
                # slice_roi=self.process_config.rec_roi,
                padding_mode=options["padding_type"],
                extra_options={
                    "scale_factor": 1.0 / options["voxel_size_cm"][0],
                    "axis_correction": options["axis_correction"],
                    "centered_axis": options["centered_axis"],
                    "clip_outer_circle": options["clip_outer_circle"],
                    "filter_cutoff": options["fbp_filter_cutoff"],
                },
            )

        # Run reconstruction
        if reconstruct_from_sinos_stack:
            # Need to copy the sino each time, as it is modified by FDK
            rec = rec_instance.reconstruct(sino.copy(), **rec_kwargs)
            # take the middle slice
            rec = rec[rec.shape[0] // 2]
        elif options["method"].lower() == "mlem" and options["implementation"].lower() == "nabu":
            rec = rec_instance(sino, n_iterations=options["iterations"]).get()
        else:
            rec = rec_instance(sino)
        rec_3D = view_as_images_stack(rec)  # writer wants 3D data

        # Write
        pipeline.writer.write_data(rec_3D)
        reconstructor.logger.info("Wrote %s" % pipeline.writer.fname)

    return 0


if __name__ == "__main__":
    main()



import os
from time import sleep
import pytest
import numpy as np
from nabu.pipeline.tasks import TaskProcessor, FakeTaskProcessor

task_processor_class = [TaskProcessor, FakeTaskProcessor]


@pytest.mark.parametrize("task_processor_class", task_processor_class)
def test_task_processor_simple(task_processor_class):
    class DummyTask:
        def __init__(self): ...

        def process_task(self, task):
            result = {
                "task_id": task["task_id"],
                "result": task["expected_result"],
            }
            return result

    task_processor = task_processor_class(DummyTask, n_processes=1)
    task_processor.start_workers()
    tasks = [{"task_id": i, "expected_result": i + 10} for i in range(5)]
    futures = []
    for task in tasks:
        future_result = task_processor.submit(0, task)
        futures.append(future_result)
    task_processor.start_work()  # not needed for TaskProcessor, but needed for FakeTaskProcessor

    # Optional: wait for results (here processing is "instant")
    for future in futures:
        result = future.result()  # blocking call
        i = result["task_id"]
        assert result["result"] == tasks[i]["expected_result"]

    task_processor.stop_workers()


@pytest.mark.parametrize("task_processor_class", task_processor_class)
def test_task_processor_callback(task_processor_class):
    class MyProcessingClass:
        def __init__(self, a):
            self.a = a
            self.previous_task = None

        def process_task(self, task):
            print(f"[{os.getpid()}] Got task: {task} (previous was: {self.previous_task})")
            sleep(0.5)
            result = f"{os.getpid()} did {task}"
            self.previous_task = task
            return result

    tp = task_processor_class(MyProcessingClass, n_processes=2, worker_init_args={"a": 1})
    tp.start_workers()

    # Submit tasks
    futures = []
    for idx, t in enumerate([f"task_{i}" for i in range(6)]):
        wid = idx % len(tp.workers)  # even/odd distribution
        fut = tp.submit(wid, t)
        fut.add_done_callback(lambda f: print(f"Callback: {f.result()}"))
        futures.append(fut)
    print("submitted")
    tp.start_work()  # not needed for TaskProcessor, but needed for FakeTaskProcessor

    # Wait for all results
    results = [f.result() for f in futures]
    print("Final results:", results)

    tp.stop_workers()


@pytest.mark.parametrize("task_processor_class", task_processor_class)
def test_with_worker_failure(task_processor_class):
    class DummyProcessing:
        def __init__(self, sleep_time):
            self.s = sleep_time

        def process_task(self, task):
            sleep(self.s)
            if task.get("fail", False):
                raise ValueError("kaboom")
            i = task.get("task_id", 1)
            return {"result": i + 10}

    tp = task_processor_class(DummyProcessing, n_processes=2, worker_init_args=(0.1,))
    tp.start_workers()

    try:
        tasks = [{"task_id": i, "fail": bool(np.random.randint(0, high=2))} for i in range(10)]
        futures = {}
        # Submit tasks, distribute evenly among workers
        for i, task in enumerate(tasks):
            w_id = i % len(tp.workers)
            f = tp.submit(w_id, task)
            futures[task["task_id"]] = f

        tp.start_work()  # not needed for TaskProcessor, but needed for FakeTaskProcessor

        # Wait for completion of all tasks
        while not (all(f.done() for f in futures.values())):
            sleep(0.5)

        # Inspect results
        for task_id, future in futures.items():
            # in principle blocking call, but since done() is True, this will return instantly
            # Also, this would hang forever if there was no failure handling mechanism
            result = future.result()
            task_failed = result.get("error", None)
            if tasks[task_id]["fail"]:
                assert task_failed is not None
            else:
                assert task_failed is None
                assert result["result"] == task_id + 10
    finally:
        tp.stop_workers()



import os
import contextlib
import h5py
import numpy as np
from silx.io.url import DataUrl
from tomoscan.io import HDF5File

from nabu.utils import first_generator_item


# This function might be moved elsewhere
def get_compacted_dataslices(urls, subsampling=None, begin=0):
    """
    Regroup urls to get the data more efficiently.
    Build a structure mapping files indices to information on
    how to load the data: `{indices_set: data_location}`
    where `data_location` contains contiguous indices.

    Parameters
    -----------
    urls: dict
        Dictionary where the key is an integer and the value is a silx `DataUrl`.
    subsampling: int, optional
        Subsampling factor when reading the frames. If an integer `n` is provided,
        then one frame out of `n` will be read.

    Returns
    --------
    merged_urls: dict
        Dictionary with the same keys as the `urls` parameter, and where the
        values are the corresponding `silx.io.url.DataUrl` with merged data_slice.
    """
    subsampling = subsampling or 1

    def _convert_to_slice(idx):
        if np.isscalar(idx):
            return slice(idx, idx + 1)
        # otherwise, assume already slice object
        return idx

    def is_contiguous_slice(slice1, slice2, step=1):
        if np.isscalar(slice1):
            slice1 = slice(slice1, slice1 + step)
        if np.isscalar(slice2):
            slice2 = slice(slice2, slice2 + step)
        return slice2.start == slice1.stop

    def merge_slices(slice1, slice2, step=1):
        return slice(slice1.start, slice2.stop, step)

    if len(urls) == 0:
        return urls

    sorted_files_indices = sorted(urls.keys())
    # if begin > 0:
    # sorted_files_indices = sorted_files_indices[begin:]
    idx0 = sorted_files_indices[begin]
    first_url = urls[idx0]

    merged_indices = [[idx0]]
    # location = (file_path, data_path, slice)
    data_location = [[first_url.file_path(), first_url.data_path(), _convert_to_slice(first_url.data_slice())]]
    pos = 0
    curr_fp, curr_dp, curr_slice = data_location[pos]
    skip_next = 0
    for idx in sorted_files_indices[begin + 1 :]:
        if skip_next > 1:
            skip_next -= 1
            continue
        url = urls[idx]
        next_slice = _convert_to_slice(url.data_slice())
        if (
            (url.file_path() == curr_fp)
            and (url.data_path() == curr_dp)
            and is_contiguous_slice(curr_slice, next_slice, step=subsampling)
        ):
            merged_indices[pos].append(idx)
            merged_slices = merge_slices(curr_slice, next_slice, step=subsampling)
            data_location[pos][-1] = merged_slices
            curr_slice = merged_slices
            skip_next = 0
        else:  # "jump"
            if begin > 0 and skip_next == 0:
                # Skip the "begin" next urls (first of a new block)
                skip_next = begin
                continue
            pos += 1
            merged_indices.append([idx])
            data_location.append([url.file_path(), url.data_path(), _convert_to_slice(url.data_slice())])
            curr_fp, curr_dp, curr_slice = data_location[pos]

    # Format result
    res = {}
    for ind, dl in zip(merged_indices, data_location):
        res.update(dict.fromkeys(ind, DataUrl(file_path=dl[0], data_path=dl[1], data_slice=dl[2])))

    return res


def get_first_hdf5_entry(fname):
    with HDF5File(fname, "r") as fid:
        entry = first_generator_item(fid.keys())
    return entry


def hdf5_entry_exists(fname, entry):
    with HDF5File(fname, "r") as fid:
        res = fid.get(entry, None) is not None
    return res


def get_h5_value(fname, h5_path, default_ret=None):
    with HDF5File(fname, "r") as fid:
        try:
            val_ptr = fid[h5_path][()]
        except KeyError:
            val_ptr = default_ret
    return val_ptr


def get_h5_str_value(dataset_ptr):
    """
    Get a HDF5 field which can be bytes or str (depending on h5py version !).
    """
    data = dataset_ptr[()]
    if isinstance(data, str):
        return data
    else:
        return bytes.decode(data)


def create_dict_of_indices(images_stack, images_indices):
    """
    From an image stack with the images indices, create a dictionary where
    each index is the image index, and the value is the corresponding image.

    Parameters
    ----------
    images_stack: numpy.ndarray
        A 3D numpy array in the layout (n_images, n_y, n_x)
    images_indices: array or list of int
        Array containing the indices of images in the stack

    Examples
    --------
    Given a simple array stack:

    >>> images_stack = np.arange(3*4*5).reshape((3,4,5))
    ... images_indices = [2, 7, 1]
    ... create_dict_of_indices(images_stack, images_indices)
    ... # returns {2: array1, 7: array2, 1: array3}
    """
    if images_stack.ndim != 3:
        raise ValueError("Expected a 3D array")
    if len(images_indices) != images_stack.shape[0]:
        raise ValueError("images_stack must have as many images as the length of images_indices")
    res = {}
    for i in range(len(images_indices)):
        res[images_indices[i]] = images_stack[i]
    return res


def convert_dict_values(dic, val_replacements, bytes_tostring=False):
    """
    Modify a dictionary to be able to export it with silx.io.dicttoh5
    """
    modified_dic = {}
    for key, value in dic.items():
        if isinstance(key, int):  # np.isscalar ?
            key = str(key)
        if isinstance(value, bytes) and bytes_tostring:
            value = bytes.decode(value.tostring())
        if isinstance(value, dict):
            value = convert_dict_values(value, val_replacements, bytes_tostring=bytes_tostring)
        else:
            if isinstance(value, DataUrl):
                value = value.path()
            elif value.__hash__ is not None and value in val_replacements:
                value = val_replacements[value]
        modified_dic[key] = value
    return modified_dic


class _BaseReader(contextlib.AbstractContextManager):
    def __init__(self, url: DataUrl):
        if not isinstance(url, DataUrl):
            raise TypeError("url should be an instance of DataUrl")
        if url.scheme() not in ("silx", "h5py"):
            raise ValueError("Valid scheme are silx and h5py")
        if url.data_slice() is not None:
            raise ValueError("Data slices are not managed. Data path should point to a bliss node (h5py.Group)")
        self._url = url
        self._file_handler = None

    def __exit__(self, *exc):
        return self._file_handler.close()


class EntryReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = HDF5File(self._url.file_path(), mode="r")
        if self._url.data_path() == "":
            entry = self._file_handler
        else:
            entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Group):
            raise TypeError("Data path should point to a bliss node (h5py.Group)")
        return entry


class DatasetReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = HDF5File(self._url.file_path(), mode="r")
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Dataset):
            raise TypeError(f"Data path ({self._url.path()}) should point to a dataset (h5py.Dataset)")
        return entry


# TODO: require some utils function to deduce type. And insure homogeneity. Might be moved in tomoscan ?
def file_format_is_edf(file_format: str):
    return file_format.lower().lstrip(".") == "edf"


def file_format_is_jp2k(file_format: str):
    return file_format.lower().lstrip(".") in ("jp2k", "jp2")


def file_format_is_tiff(file_format: str):
    return file_format.lower().lstrip(".") in ("tiff", "tif")


def file_format_is_hdf5(file_format: str):
    return file_format.lower().lstrip(".") in ("hdf5", "hdf", "nx", "nexus")


def get_output_volume(location: str, file_prefix: str | None, file_format: str, multitiff=False):

    # FIXME
    # These imports take an awful lot of time.
    # Waiting for the fix in tomoscan & underlying libs, the imports are done here for now
    from tomoscan.esrf import EDFVolume, HDF5Volume, TIFFVolume, JP2KVolume, MultiTIFFVolume

    #

    # TODO: see strategy. what if user provide a .nx ... ?
    # this function should be more generic
    location, extension = os.path.splitext(location)
    if extension == "":
        extension = file_format
    if file_format_is_edf(extension):
        return EDFVolume(folder=location, volume_basename=file_prefix)
    elif file_format_is_jp2k(extension):
        return JP2KVolume(folder=location, volume_basename=file_prefix)
    elif file_format_is_hdf5(file_format=extension):
        if extension is None:
            if file_prefix is None:
                location = ".".join([location, extension])
            else:
                location = os.path.join(location, ".".join([file_prefix, extension]))
        return HDF5Volume(file_path=location)
    elif file_format_is_tiff(extension):
        if multitiff:
            return MultiTIFFVolume(file_path=location)
        else:
            return TIFFVolume(folder=location, volume_basename=file_prefix)
    else:
        raise ValueError



import pytest
import numpy as np
import scipy.interpolate
from nabu.processing.fft_cuda import get_available_fft_implems
from nabu.testutils import get_data as nabu_get_data
from nabu.testutils import __do_long_tests__
from nabu.preproc.flatfield import FlatFieldArrays
from nabu.preproc.ccd import CCDFilter
from nabu.preproc import ctf
from nabu.estimation.distortion import estimate_flat_distortion
from nabu.misc.filters import correct_spikes
from nabu.preproc.distortion import DistortionCorrection
from nabu.cuda.utils import __has_cupy__

__has_cufft__ = False
if __has_cupy__:
    from nabu.preproc.ctf_cuda import CudaCTFPhaseRetrieval

    avail_fft = get_available_fft_implems()
    __has_cufft__ = len(avail_fft) > 0


@pytest.fixture(scope="class")
def bootstrap_TestCtf(request):
    cls = request.cls
    cls.abs_tol = 1.0e-4

    test_data = nabu_get_data("ctf_tests_data_all_pars.npz")

    cls.rand_disp_vh = test_data["rh"]
    ## the dimension number 1 is over holotomo distances, so far our filter is for one distance only
    cls.rand_disp_vh.shape = [cls.rand_disp_vh.shape[0], cls.rand_disp_vh.shape[2]]
    cls.dark = test_data["dark"]
    cls.flats = [test_data["ref0"], test_data["ref1"]]
    cls.im = test_data["im"]
    cls.ipro = int(test_data["ipro"][0])
    cls.expected_result = test_data["result"]
    cls.ref_plain = test_data["ref_plain_float_flat"]

    cls.flats_n = test_data["refns"]

    cls.img_shape_vh = test_data["img_shape_vh"]
    cls.padded_img_shape_vh = test_data["padded_img_shape_vh"]
    cls.z1_vh = test_data["z1_vh"]
    cls.z2 = test_data["z2"]
    cls.pix_size_det = test_data["pix_size_det"][()]
    cls.length_scale = test_data["length_scale"]
    cls.wavelength = test_data["wave_length"]
    cls.remove_spikes_threshold = test_data["remove_spikes_threshold"]
    cls.delta_beta = 27


@pytest.mark.usefixtures("bootstrap_TestCtf")
class TestCtf:
    def check_result(self, res, ref, error_message):
        diff = np.abs(res - ref)
        diff[diff > np.percentile(diff, 99)] = 0
        assert diff.max() < self.abs_tol * (np.abs(ref).mean()), error_message

    def test_ctf_id16_way(self):
        """Test the CTF phase retrieval.
        The CTF filter, from the CtfFilter class, is initialised with the geometry information contained in the GeoPars object.
        The geometry encompasses the case of an astigmatic wavefront with vertical and horizontal sources located at
        distances z1_vh[0] and z1_vh[1] from the object.
        In the case of parallel geometry, set z1_vh[0] = z1_vh[1] = R where R is a large value (meters).
        The SI unit system is used, but the same results should be obtained with any homogeneous choice of distance units.
        The `img_shape` is the shape of the images to be processed.
        `padded_img_shape` is an intermediate shape that must be larger than `img_shape` to avoid border effects due
        to convolutions.
        `length_scale` is an internal parameter that should not affect the result unless there are serious numerical
        problems involving very small lengths. You can safely keep the default value.
        """
        geo_pars = ctf.GeoPars(
            z1_vh=self.z1_vh,
            z2=self.z2,
            pix_size_det=self.pix_size_det,
            length_scale=self.length_scale,
            wavelength=self.wavelength,
        )

        flats = FlatFieldArrays(
            [1200] + list(self.img_shape_vh), {0: self.flats[0], 1200: self.flats[1]}, {0: self.dark}
        )

        my_flat = flats.get_flat(self.ipro)
        my_img = self.im - self.dark
        my_flat = my_flat - self.dark

        new_coordinates = estimate_flat_distortion(
            my_flat,
            my_img,
            tile_size=100,
            interpolation_kind="cubic",
            padding_mode="edge",
            correction_spike_threshold=3,
        )

        interpolator = scipy.interpolate.RegularGridInterpolator(
            (np.arange(my_flat.shape[0]), np.arange(my_flat.shape[1])),
            my_flat,
            bounds_error=False,
            method="linear",
            fill_value=None,
        )
        my_flat = interpolator(new_coordinates)

        my_img = my_img / my_flat
        my_img = correct_spikes(my_img, self.remove_spikes_threshold)

        my_shift = self.rand_disp_vh[:, self.ipro]

        ctf_filter = ctf.CtfFilter(
            self.dark.shape,
            geo_pars,
            self.delta_beta,
            padded_shape=self.padded_img_shape_vh,
            translation_vh=my_shift,
            normalize_by_mean=True,
            lim1=1.0e-5,
            lim2=0.2,
        )
        phase = ctf_filter.retrieve_phase(my_img)

        self.check_result(
            phase, self.expected_result, "retrieved phase and reference result differ beyond the accepted tolerance"
        )

    @pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
    def test_ctf_id16_class(self):
        geo_pars = ctf.GeoPars(
            z1_vh=self.z1_vh,
            z2=self.z2,
            pix_size_det=self.pix_size_det,
            length_scale=self.length_scale,
            wavelength=self.wavelength,
        )
        distortion_correction = DistortionCorrection(
            estimation_method="fft-correlation",
            estimation_kwargs={
                "tile_size": 100,
                "interpolation_kind": "cubic",
                "padding_mode": "edge",
                "correction_spike_threshold": 3.0,
            },
            correction_method="interpn",
            correction_kwargs={"fill_value": None},
        )
        flats = FlatFieldArrays(
            [1200] + list(self.img_shape_vh),
            {0: self.flats[0], 1200: self.flats[1]},
            {0: self.dark},
            distortion_correction=distortion_correction,
        )

        # The "correct_spikes" function is numerically unstable (comparison with a float threshold).
        # If float32 is used for the image, one spike is detected while it is not in the previous test
        # (although the max difference between the inputs is about 1e-8).
        # We use float64 data type for the image to make tests pass.
        img = self.im.astype(np.float64)

        flats.normalize_single_radio(img, self.ipro)
        img = correct_spikes(img, self.remove_spikes_threshold)

        shift = self.rand_disp_vh[:, self.ipro]

        ctf_filter = ctf.CtfFilter(
            img.shape,
            geo_pars,
            self.delta_beta,
            padded_shape=self.padded_img_shape_vh,
            translation_vh=shift,
            normalize_by_mean=True,
            lim1=1.0e-5,
            lim2=0.2,
        )
        phase = ctf_filter.retrieve_phase(img)

        message = "retrieved phase and reference result differ beyond the accepted tolerance"
        assert np.abs(phase - self.expected_result).max() < 10 * self.abs_tol * (
            np.abs(self.expected_result).mean()
        ), message

    def test_ctf_plain_way(self):
        geo_pars = ctf.GeoPars(
            z1_vh=None,
            z2=self.z2,
            pix_size_det=self.pix_size_det,
            length_scale=self.length_scale,
            wavelength=self.wavelength,
        )
        flatfielder = FlatFieldArrays(
            [1] + list(self.img_shape_vh),
            {0: self.flats[0], 1200: self.flats[1]},
            {0: self.dark},
            radios_indices=[self.ipro],
        )

        spikes_corrector = CCDFilter(
            self.dark.shape, median_clip_thresh=self.remove_spikes_threshold, abs_diff=True, preserve_borders=True
        )

        img = self.im.astype("f")
        img = flatfielder.normalize_radios(np.array([img]))[0]
        img = spikes_corrector.median_clip_correction(img)

        ctf_args = [img.shape, geo_pars, self.delta_beta]
        ctf_kwargs = {"padded_shape": self.padded_img_shape_vh, "normalize_by_mean": True, "lim1": 1.0e-5, "lim2": 0.2}
        ctf_filter = ctf.CtfFilter(*ctf_args, **ctf_kwargs)
        phase = ctf_filter.retrieve_phase(img)

        self.check_result(phase, self.ref_plain, "Something wrong with CtfFilter")

        # Test R2C
        ctf_numpy = ctf.CtfFilter(*ctf_args, **ctf_kwargs, use_rfft=True)
        phase_r2c = ctf_numpy.retrieve_phase(img)
        self.check_result(phase_r2c, self.ref_plain, "Something wrong with CtfFilter-R2C")

        # Test multi-core FFT
        ctf_fft = ctf.CtfFilter(*ctf_args, **ctf_kwargs, use_rfft=True, fft_num_threads=0)
        if ctf_fft.use_rfft:
            # phase_fft = ctf_fft.retrieve_phase(img)
            self.check_result(phase_r2c, self.ref_plain, "Something wrong with CtfFilter-FFT")

    @pytest.mark.skipif(not (__has_cupy__ and __has_cufft__), reason="need cupy and vkfft")
    def test_cuda_ctf(self):
        data = nabu_get_data("brain_phantom.npz")["data"]
        delta_beta = 50.0
        energy_kev = 22.0
        distance_m = 1.0
        pix_size_m = 0.1e-6

        geo_pars = ctf.GeoPars(z2=distance_m, pix_size_det=pix_size_m, wavelength=1.23984199e-9 / energy_kev)

        for normalize in [True, False]:
            ctf_filter = ctf.CTFPhaseRetrieval(
                data.shape, geo_pars, delta_beta=delta_beta, normalize_by_mean=normalize, use_rfft=True
            )
            cuda_ctf_filter = CudaCTFPhaseRetrieval(
                data.shape,
                geo_pars,
                delta_beta=delta_beta,
                use_rfft=True,
                normalize_by_mean=normalize,
            )
            ref = ctf_filter.retrieve_phase(data)

            d_data = cuda_ctf_filter.cuda_processing.to_device("_d_data", data)
            res = cuda_ctf_filter.retrieve_phase(d_data).get()
            err_max = np.max(np.abs(res - ref))

            assert err_max < 1e-2, "Something wrong with retrieve_phase(normalize_by_mean=%s)" % (str(normalize))



"""
nabu.pipeline.estimators: helper classes/functions to estimate parameters of a dataset
(center of rotation, detector tilt, etc).
"""

import inspect
import numpy as np
import scipy.fft  #  pylint: disable=E0611
from silx.io import get_data
import math
from scipy import ndimage as nd

from ..preproc.flatfield import FlatField
from ..estimation.cor import (
    CenterOfRotation,
    CenterOfRotationAdaptiveSearch,
    CenterOfRotationSlidingWindow,
    CenterOfRotationGrowingWindow,
    CenterOfRotationOctaveAccurate,
)
from .. import version as nabu_version
from ..estimation.cor_sino import SinoCorInterface, CenterOfRotationFourierAngles, CenterOfRotationVo
from ..estimation.tilt import CameraTilt
from ..estimation.motion import MotionEstimation
from ..estimation.utils import is_fullturn_scan
from ..resources.logger import LoggerOrPrint
from ..resources.utils import extract_parameters
from ..utils import check_supported, deprecation_warning, get_num_threads, is_int, is_scalar
from ..resources.dataset_analyzer import get_radio_pair
from ..processing.rotation import Rotation
from ..preproc.ccd import Log, CCDFilter
from ..misc import fourier_filters
from .params import cor_methods, tilt_methods


def estimate_cor(method, dataset_info, do_flatfield=True, cor_options=None, logger=None):
    """
    High level function to compute the center of rotation (COR)

    Parameters
    ----------
    method: name of the method to be used for computing the center of rotation
    dataset_info: `nabu.resources.dataset_analyzer.DatasetAnalyzer`
        Dataset information structure
    do_flatfield: If True apply flat field to compute the center of rotation
    cor_options: optional dictionary that can contain the following keys:
        * slice_idx: index of the slice to use for computing the sinogram (for sinogram based algorithms)
        * subsampling subsampling
        * radio_angles: angles of the radios to use (for radio based algorithms)
    logger: logging object
    """
    logger = LoggerOrPrint(logger)
    cor_options = cor_options or {}
    check_supported(method, list(cor_methods.keys()), "COR estimation method")
    method = cor_methods[method]

    # Extract CoR parameters from configuration file
    if isinstance(cor_options, str):
        try:
            cor_options = extract_parameters(cor_options, sep=";")
        except Exception as exc:
            msg = "Could not extract parameters from cor_options: %s" % (str(exc))
            logger.fatal(msg)
            raise ValueError(msg)
    elif isinstance(cor_options, dict):
        pass
    else:
        raise TypeError(f"cor_options_str is expected to be a dict or a str. {type(cor_options)} provided")

    # Dispatch. COR estimation is always expressed in absolute number of pixels (i.e. from the center of the first pixel column)
    if method in CORFinder.search_methods:
        cor_finder = CORFinder(
            method,
            dataset_info,
            do_flatfield=do_flatfield,
            cor_options=cor_options,
            radio_angles=cor_options.get("radio_angles", (0.0, np.pi)),
            logger=logger,
        )
        estimated_cor = cor_finder.find_cor()
    elif method in SinoCORFinder.search_methods:
        cor_finder = SinoCORFinder(
            method,
            dataset_info,
            slice_idx=cor_options.get("slice_idx", "middle"),
            subsampling=cor_options.get("subsampling", 10),
            do_flatfield=do_flatfield,
            take_log=cor_options.get("take_log", True),
            cor_options=cor_options,
            logger=logger,
        )
        estimated_cor = cor_finder.find_cor()
    else:
        composite_options = update_func_kwargs(CompositeCORFinder, cor_options)
        for what in ["cor_options", "logger", "do_flatfield"]:
            composite_options.pop(what, None)
        cor_finder = CompositeCORFinder(
            dataset_info,
            cor_options=cor_options,
            do_flatfield=do_flatfield,
            logger=logger,
            **composite_options,  # contains take_log
        )
        estimated_cor = cor_finder.find_cor()
    return estimated_cor


class CORFinderBase:
    """
    A base class for CoR estimators.
    It does common tasks like data reading, flatfield, etc.
    """

    search_methods = {}

    def __init__(self, method, dataset_info, do_flatfield=True, cor_options=None, logger=None):
        """
        Initialize a CORFinder object.

        Parameters
        ----------
        dataset_info: `nabu.resources.dataset_analyzer.DatasetAnalyzer`
            Dataset information structure
        """
        check_supported(method, self.search_methods, "CoR estimation method")
        self.method = method
        self.cor_options = cor_options or {}
        self.logger = LoggerOrPrint(logger)
        self.dataset_info = dataset_info
        self.do_flatfield = do_flatfield
        self.shape = dataset_info.radio_dims[::-1]
        self._get_lookup_side()
        self._init_cor_finder()

    def _get_lookup_side(self):
        """
        Get the "initial guess" where the center-of-rotation (CoR) should be estimated.
        For example 'center' means that CoR search will be done near the middle of the detector, i.e center column.
        """
        lookup_side = self.cor_options.get("side", None)
        self._lookup_side = lookup_side
        # User-provided scalar
        if not (isinstance(lookup_side, str)) and np.isscalar(lookup_side):
            return

        default_lookup_side = "right" if self.dataset_info.is_halftomo else "center"

        # By default in nabu config, side='from_file' meaning that we inspect the dataset information for CoR metadata
        if lookup_side == "from_file":
            initial_cor_pos = self.dataset_info.dataset_scanner.x_rotation_axis_pixel_position  # relative pos in pixels
            if initial_cor_pos is None:
                self.logger.warning("Could not get an initial estimate for center of rotation in data file")
                lookup_side = default_lookup_side
            else:
                lookup_side = initial_cor_pos
            self._lookup_side = lookup_side

    def _init_cor_finder(self):
        cor_finder_cls = self.search_methods[self.method]["class"]
        self.cor_finder = cor_finder_cls(verbose=False, logger=self.logger, extra_options=None)


class CORFinder(CORFinderBase):
    """
    Find the Center of Rotation with methods based on two (180-degrees opposed) radios.
    """

    search_methods = {
        "centered": {
            "class": CenterOfRotation,
        },
        "global": {
            "class": CenterOfRotationAdaptiveSearch,
            "default_kwargs": {"low_pass": 1, "high_pass": 20},
        },
        "sliding-window": {
            "class": CenterOfRotationSlidingWindow,
        },
        "growing-window": {
            "class": CenterOfRotationGrowingWindow,
        },
        "octave-accurate": {
            "class": CenterOfRotationOctaveAccurate,
        },
    }

    def __init__(
        self, method, dataset_info, do_flatfield=True, cor_options=None, logger=None, radio_angles=(0.0, np.pi)
    ):
        """
        Initialize a CORFinder object.

        Parameters
        ----------
        dataset_info: `nabu.resources.dataset_analyzer.DatasetAnalyzer`
            Dataset information structure
        radio_angles: angles to use to find the cor
        """
        super().__init__(method, dataset_info, do_flatfield=do_flatfield, cor_options=cor_options, logger=logger)
        self._radio_angles = radio_angles
        self._init_radios()
        self._apply_flatfield()
        self._apply_tilt()
        # octave-accurate does not support half-acquisition scans,
        # but information on field of view is only known here with the "dataset_info" object.
        # Do the check here.
        if self.dataset_info.is_halftomo and method == "octave-accurate":
            raise ValueError("The CoR estimator 'octave-accurate' does not support half-acquisition scans")
        #

    def _init_radios(self):
        self.radios, self._radios_indices = get_radio_pair(
            self.dataset_info, radio_angles=self._radio_angles, return_indices=True
        )

    def _apply_flatfield(self):
        if not (self.do_flatfield):
            return
        self.flatfield = FlatField(
            self.radios.shape,
            flats=self.dataset_info.flats,
            darks=self.dataset_info.darks,
            radios_indices=self._radios_indices,
            interpolation="linear",
        )
        self.flatfield.normalize_radios(self.radios)

    def _apply_tilt(self):
        tilt = self.dataset_info.detector_tilt
        if tilt is None:
            return
        self.logger.debug("COREstimator: applying detector tilt correction of %f degrees" % tilt)
        rot = Rotation(self.shape, tilt)
        for i in range(self.radios.shape[0]):
            self.radios[i] = rot.rotate(self.radios[i])

    def find_cor(self):
        """
        Find the center of rotation.

        Returns
        -------
        cor: float
            The estimated center of rotation for the current dataset.
        """
        self.logger.info("Estimating center of rotation")
        # All find_shift() methods in self.search_methods have the same API with "img_1" and "img_2"
        cor_exec_kwargs = update_func_kwargs(self.cor_finder.find_shift, self.cor_options)
        cor_exec_kwargs["return_relative_to_middle"] = False
        # ----- FIXME -----
        # 'self.cor_options' can contain 'side="from_file"', and we should not modify it directly
        # because it's entered by the user.
        # Either make a copy of self.cor_options, or change the inspect() mechanism
        if cor_exec_kwargs.get("side", None) == "from_file":
            cor_exec_kwargs["side"] = self._lookup_side or "center"
        # ------
        if self._lookup_side is not None:
            cor_exec_kwargs["side"] = self._lookup_side
        self.logger.debug("%s.find_shift(%s)" % (self.cor_finder.__class__.__name__, str(cor_exec_kwargs)))
        shift = self.cor_finder.find_shift(self.radios[0], np.fliplr(self.radios[1]), **cor_exec_kwargs)
        return shift


# alias
COREstimator = CORFinder


class SinoCORFinder(CORFinderBase):
    """
    A class for finding Center of Rotation based on 360 degrees sinograms.
    This class handles the steps of building the sinogram from raw radios.
    """

    search_methods = {
        "sino-coarse-to-fine": {
            "class": SinoCorInterface,
        },
        "sino-sliding-window": {
            "class": CenterOfRotationSlidingWindow,
        },
        "sino-growing-window": {
            "class": CenterOfRotationGrowingWindow,
        },
        "fourier-angles": {"class": CenterOfRotationFourierAngles},
        "vo": {
            "class": CenterOfRotationVo,
        },
    }

    def __init__(
        self,
        method,
        dataset_info,
        do_flatfield=True,
        take_log=True,
        cor_options=None,
        logger=None,
        slice_idx="middle",
        subsampling=10,
    ):
        """
        Initialize a SinoCORFinder object.

        Other Parameters
        ----------------
        The following keys can be set in cor_options.

        slice_idx: int or str
            Which slice index to take for building the sinogram.
            For example slice_idx=0 means that we extract the first line of each projection.
            Value can also be "first", "top", "middle", "last", "bottom".
        subsampling: int, float
            subsampling strategy when building sinograms.
            As building the complete sinogram from raw projections might be tedious, the reading is done with subsampling.
            A positive integer value means the subsampling step (i.e `projections[::subsampling]`).
        """
        super().__init__(method, dataset_info, do_flatfield=do_flatfield, cor_options=cor_options, logger=logger)
        self._set_slice_idx(slice_idx)
        self._set_subsampling(subsampling)
        self._load_raw_sinogram()
        self._flatfield(do_flatfield)
        self._get_sinogram(take_log)

    def _check_360(self):
        if not is_fullturn_scan(self.dataset_info.rotation_angles):
            raise ValueError("Sinogram-based Center of Rotation estimation can only be used for 360 degrees scans")

    def _set_slice_idx(self, slice_idx):
        n_z = self.dataset_info.radio_dims[1]
        if isinstance(slice_idx, str):
            str_to_idx = {"top": 0, "first": 0, "middle": n_z // 2, "bottom": n_z - 1, "last": n_z - 1}
            check_supported(slice_idx, str_to_idx.keys(), "slice location")
            slice_idx = str_to_idx[slice_idx]
        self.slice_idx = slice_idx

    def _set_subsampling(self, subsampling):
        projs_idx = sorted(self.dataset_info.projections.keys())
        self.subsampling = None
        if is_int(subsampling):
            if subsampling < 0:  # Total number of angles
                raise NotImplementedError
            else:
                self.projs_indices = projs_idx[::subsampling]
                self.angles = self.dataset_info.rotation_angles[::subsampling]
                self.subsampling = subsampling
        else:  # Angular step
            raise NotImplementedError

    def _load_raw_sinogram(self):
        if self.slice_idx is None:
            raise ValueError("Unknow slice index")
        reader_kwargs = {
            "sub_region": (slice(None, None, self.subsampling), slice(self.slice_idx, self.slice_idx + 1), slice(None))
        }
        if self.dataset_info.kind == "edf":
            reader_kwargs = {"n_reading_threads": get_num_threads()}
        self.data_reader = self.dataset_info.get_reader(**reader_kwargs)
        self._radios = self.data_reader.load_data()

    def _flatfield(self, do_flatfield):
        self.do_flatfield = bool(do_flatfield)
        if not self.do_flatfield:
            return
        flats = {k: arr[self.slice_idx : self.slice_idx + 1, :] for k, arr in self.dataset_info.flats.items()}
        darks = {k: arr[self.slice_idx : self.slice_idx + 1, :] for k, arr in self.dataset_info.darks.items()}
        flatfield = FlatField(
            self._radios.shape,
            flats,
            darks,
            radios_indices=self.projs_indices,
        )
        flatfield.normalize_radios(self._radios)

    def _get_sinogram(self, take_log):
        sinogram = self._radios[:, 0, :].copy()
        if take_log:
            log = Log(self._radios.shape, clip_min=1e-6, clip_max=10.0)
            log.take_logarithm(sinogram)
        self.sinogram = sinogram

    @staticmethod
    def _split_sinogram(sinogram):
        n_a_2 = sinogram.shape[0] // 2
        img_1, img_2 = sinogram[:n_a_2], sinogram[n_a_2:]
        # "Handle" odd number of projections
        if img_2.shape[0] > img_1.shape[0]:
            img_2 = img_2[:-1, :]
        #
        return img_1, img_2

    def find_cor(self):
        self.logger.info("Estimating center of rotation")

        cor_exec_kwargs = update_func_kwargs(self.cor_finder.find_shift, self.cor_options)
        cor_exec_kwargs["return_relative_to_middle"] = False
        # FIXME
        # 'self.cor_options' can contain 'side="from_file"', and we should not modify it directly
        # because it's entered by the user.
        # Either make a copy of self.cor_options, or change the inspect() mechanism
        if cor_exec_kwargs["side"] == "from_file":
            cor_exec_kwargs["side"] = self._lookup_side or "center"
        #
        if self._lookup_side is not None:
            cor_exec_kwargs["side"] = self._lookup_side

        if self.method == "fourier-angles":
            cor_exec_args = [self.sinogram]
            cor_exec_kwargs["angles"] = self.dataset_info.rotation_angles
        elif self.method == "vo":
            cor_exec_args = [self.sinogram]
            cor_exec_kwargs["halftomo"] = self.dataset_info.is_halftomo
            cor_exec_kwargs["is_360"] = is_fullturn_scan(self.dataset_info.rotation_angles)
        else:
            # For these methods relying on find_shift() with two images, the sinogram needs to be split in two
            img_1, img_2 = self._split_sinogram(self.sinogram)
            cor_exec_args = [img_1, np.fliplr(img_2)]

        self.logger.debug("%s.find_shift(%s)" % (self.cor_finder.__class__.__name__, str(cor_exec_kwargs)))
        shift = self.cor_finder.find_shift(*cor_exec_args, **cor_exec_kwargs)

        return shift


# alias
SinoCOREstimator = SinoCORFinder


class CompositeCORFinder(CORFinderBase):
    """
    Class and method to prepare sinogram and calculate COR
    The pseudo sinogram is built with shrunken radios taken every theta_interval degrees

    Compared to first writing by Christian Nemoz:
        - gives the same result of the original octave script on the dataset so far tested
        - The meaning of parameter n_subsampling_y (alias subsampling_y)is now the number of lines which are taken from
          every radio. This is more meaningful in terms of amount of collected information because it
          does not depend on the radio size. Moreover this is what was done in the octave script
        - The spike_threshold has been added with default to 0.04
        - The angular sampling is every 5 degree by default, as it is now the case also in the octave script
        - The finding of the optimal overlap is doing by looping over the possible overlap, according to the overlap.
           After a first testing phase, this part, which is the time consuming part, can be accelerated
           by several order of magnitude without modifying the final result
    """

    search_methods = {
        "composite-coarse-to-fine": {
            "class": CenterOfRotation,  # Not used. Everything is done in the find_cor() func.
        }
    }
    _default_cor_options = {"low_pass": 0.4, "high_pass": 10, "side": "near", "near_pos": 0, "near_width": 100}

    def __init__(
        self,
        dataset_info,
        oversampling=4,
        theta_interval=5,
        n_subsampling_y=40,
        do_flatfield=True,
        take_log=True,
        cor_options=None,
        spike_threshold=0.04,
        logger=None,
        norm_order=1,
    ):
        super().__init__(
            "composite-coarse-to-fine", dataset_info, do_flatfield=False, cor_options=cor_options, logger=logger
        )
        if norm_order not in [1, 2]:
            raise ValueError(
                f""" the norm order (nom_order parameter) must be either 1 or 2. You passed {norm_order}
                """
            )

        self.norm_order = norm_order

        self.dataset_info = dataset_info
        self.logger = LoggerOrPrint(logger)

        self.sx, self.sy = self.dataset_info.radio_dims

        default_cor_options = self._default_cor_options.copy()
        default_cor_options.update(self.cor_options)
        self.cor_options = default_cor_options

        # the algorithm can work for angular ranges larger than 1.2*pi
        # up to an arbitrarily number of turns as it is the case in helical scans
        self.spike_threshold = spike_threshold
        # the following line is necessary for multi-turns scan because the encoders is always
        # in the interval 0-360
        self.unwrapped_rotation_angles = np.unwrap(self.dataset_info.rotation_angles)

        self.angle_min = self.unwrapped_rotation_angles.min()
        self.angle_max = self.unwrapped_rotation_angles.max()

        if (self.angle_max - self.angle_min) < 1.2 * np.pi:
            useful_span = None
            raise ValueError(
                f"""Sinogram-based Center of Rotation estimation can only be used for scans over more than 180 degrees.
                                 Your angular span was barely above 180 degrees, it was in fact {((self.angle_max - self.angle_min)/np.pi):.2f} x 180
                                 and it is not considered to be enough by the discriminating condition which requires at least 1.2 half-turns
                              """
            )
        else:
            useful_span = min(np.pi, (self.angle_max - self.angle_min) - np.pi)
            # readapt theta_interval accordingly if the span is smaller than pi
            if useful_span < np.pi:
                theta_interval = theta_interval * useful_span / np.pi

        self.take_log = take_log
        self.ovs = oversampling
        self.theta_interval = theta_interval

        target_sampling_y = np.round(np.linspace(0, self.sy - 1, n_subsampling_y + 2)).astype(int)[1:-1]

        if self.spike_threshold is not None:
            # take also one line below and on above for each line
            # to provide appropriate margin
            self.sampling_y = np.zeros([3 * len(target_sampling_y)], "i")
            self.sampling_y[0::3] = np.maximum(0, target_sampling_y - 1)
            self.sampling_y[2::3] = np.minimum(self.sy - 1, target_sampling_y + 1)
            self.sampling_y[1::3] = target_sampling_y

            self.ccd_correction = CCDFilter((len(self.sampling_y), self.sx), median_clip_thresh=self.spike_threshold)
        else:
            self.sampling_y = target_sampling_y

        self.nproj = self.dataset_info.n_angles

        my_condition = np.less(self.unwrapped_rotation_angles + np.pi, self.angle_max) * np.less(
            self.unwrapped_rotation_angles, self.angle_min + useful_span
        )

        possibly_probed_angles = self.unwrapped_rotation_angles[my_condition]
        possibly_probed_indices = np.arange(len(self.unwrapped_rotation_angles))[my_condition]

        self.dproj = round(len(possibly_probed_angles) / np.rad2deg(useful_span) * self.theta_interval)

        self.probed_angles = possibly_probed_angles[:: self.dproj]
        self.probed_indices = possibly_probed_indices[:: self.dproj]

        self.absolute_indices = sorted(self.dataset_info.projections.keys())

        if do_flatfield:
            my_flats = self.dataset_info.flats
        else:
            my_flats = None

        if my_flats is not None and len(list(my_flats.keys())) > 0:
            self.use_flat = True
            self.flatfield = FlatField(
                (len(self.absolute_indices), self.sy, self.sx),
                self.dataset_info.flats,
                self.dataset_info.darks,
                radios_indices=self.absolute_indices,
            )
        else:
            self.use_flat = False

        self.sx, self.sy = self.dataset_info.radio_dims
        self.mlog = Log((1,) + (self.sy, self.sx), clip_min=1e-6, clip_max=10.0)
        self.rcor_abs = round(self.sx / 2.0)
        self.cor_acc = round(self.sx / 2.0)

        self.nprobed = len(self.probed_angles)

        # initialize sinograms and radios arrays
        self.sino = np.zeros([2 * self.nprobed * n_subsampling_y, (self.sx - 1) * self.ovs + 1], "f")
        self._loaded = False
        self.high_pass = self.cor_options["high_pass"]
        img_filter = fourier_filters.get_bandpass_filter(
            (self.sino.shape[0] // 2, self.sino.shape[1]),
            cutoff_lowpass=self.cor_options["low_pass"] * self.ovs,
            cutoff_highpass=self.high_pass * self.ovs,
            use_rfft=False,  # rfft changes the image dimensions lenghts to even if odd
            data_type=np.float64,
        )

        # we are interested in filtering only along the x dimension only
        img_filter[:] = img_filter[0]
        self.img_filter = img_filter

    def _oversample(self, radio):
        """oversampling in the horizontal direction"""
        if self.ovs == 1:
            return radio
        else:
            ovs_2D = [1, self.ovs]
        return oversample(radio, ovs_2D)

    def _get_cor_options(self, cor_options):
        default_dict = self._default_cor_options.copy()
        if self.dataset_info.is_halftomo:
            default_dict["side"] = "right"

        if cor_options is None or cor_options == "":
            cor_options = {}
        if isinstance(cor_options, str):
            try:
                cor_options = extract_parameters(cor_options, sep=";")
            except Exception as exc:
                msg = "Could not extract parameters from cor_options: %s" % (str(exc))
                self.logger.fatal(msg)
                raise ValueError(msg)
        default_dict.update(cor_options)
        cor_options = default_dict

        self.cor_options = cor_options

    def get_radio(self, image_num):
        #  radio_dataset_idx = self.absolute_indices[image_num]
        radio_dataset_idx = image_num
        data_url = self.dataset_info.projections[radio_dataset_idx]
        radio = get_data(data_url).astype(np.float64)
        if self.use_flat:
            self.flatfield.normalize_single_radio(radio, radio_dataset_idx, dtype=radio.dtype)
        if self.take_log:
            self.mlog.take_logarithm(radio)

        radio = radio[self.sampling_y]
        if self.spike_threshold is not None:
            self.ccd_correction.median_clip_correction(radio, output=radio)
            radio = radio[1::3]
        return radio

    def get_sino(self, reload=False):
        """
        Build sinogram (composite image) from the radio files
        """
        if self._loaded and not reload:
            return self.sino

        sorting_indexes = np.argsort(self.unwrapped_rotation_angles)

        sorted_all_angles = self.unwrapped_rotation_angles[sorting_indexes]
        sorted_angle_indexes = np.arange(len(self.unwrapped_rotation_angles))[sorting_indexes]

        irad = 0
        for prob_a, prob_i in zip(self.probed_angles, self.probed_indices):
            radio1 = self.get_radio(self.absolute_indices[prob_i])
            other_angle = prob_a + np.pi

            insertion_point = np.searchsorted(sorted_all_angles, other_angle)
            if insertion_point > 0 and insertion_point < len(sorted_all_angles):
                other_i_l = sorted_angle_indexes[insertion_point - 1]
                other_i_h = sorted_angle_indexes[insertion_point]
                radio_l = self.get_radio(self.absolute_indices[other_i_l])
                radio_h = self.get_radio(self.absolute_indices[other_i_h])
                f = (other_angle - sorted_all_angles[insertion_point - 1]) / (
                    sorted_all_angles[insertion_point] - sorted_all_angles[insertion_point - 1]
                )
                radio2 = (1 - f) * radio_l + f * radio_h
            else:
                if insertion_point == 0:
                    other_i = sorted_angle_indexes[0]
                elif insertion_point == len(sorted_all_angles):
                    other_i = sorted_angle_indexes[insertion_point - 1]
                radio2 = self.get_radio(self.absolute_indices[other_i])  # pylint: disable=E0606

            self.sino[irad : irad + radio1.shape[0], :] = self._oversample(radio1)
            self.sino[
                irad + self.nprobed * radio1.shape[0] : irad + self.nprobed * radio1.shape[0] + radio1.shape[0], :
            ] = self._oversample(radio2)

            irad = irad + radio1.shape[0]

        self.sino[np.isnan(self.sino)] = 0.0001  # ?
        return self.sino

    def find_cor(self, reload=False):
        self.logger.info("Estimating center of rotation")
        self.logger.debug("%s.find_shift(%s)" % (self.__class__.__name__, self.cor_options))
        self.sinogram = self.get_sino(reload=reload)

        detector_center = (self.shape[-1] - 1) / 2.0
        dim_v, dim_h = self.sinogram.shape
        assert dim_v % 2 == 0, " this should not happen "
        dim_v = dim_v // 2

        radio1 = self.sinogram[:dim_v]
        radio2 = self.sinogram[dim_v:]

        orig_sy, orig_ovsd_sx = radio1.shape

        radio1 = scipy.fft.ifftn(
            scipy.fft.fftn(radio1, axes=(-2, -1)) * self.img_filter, axes=(-2, -1)
        ).real  # TODO: convolute only along x
        radio2 = scipy.fft.ifftn(
            scipy.fft.fftn(radio2, axes=(-2, -1)) * self.img_filter, axes=(-2, -1)
        ).real  # TODO: convolute only along x

        tmp_sy, radio_width = radio1.shape
        assert orig_sy == tmp_sy and orig_ovsd_sx == radio_width, "this should not happen"

        cor_side = self._lookup_side if self._lookup_side is not None else self.cor_options["side"]
        if cor_side == "center":
            overlap_min = max(round(radio_width - radio_width / 3), 4)
            overlap_max = min(round(radio_width + radio_width / 3), 2 * radio_width - 4)
        elif cor_side == "right":
            overlap_min = max(4, self.ovs * self.high_pass * 3)
            overlap_max = radio_width
        elif cor_side == "left":
            overlap_min = radio_width
            overlap_max = min(2 * radio_width - 4, 2 * radio_width - self.ovs * self.ovs * self.high_pass * 3)
        elif cor_side == "all":
            overlap_min = max(4, self.ovs * self.high_pass * 3)
            overlap_max = min(2 * radio_width - 4, 2 * radio_width - self.ovs * self.ovs * self.high_pass * 3)
        elif is_scalar(cor_side):
            near_pos = cor_side
            near_width = self.cor_options["near_width"]
            overlap_min = max(4, radio_width - 2 * self.ovs * (near_pos + near_width))
            overlap_max = min(2 * radio_width - 4, radio_width - 2 * self.ovs * (near_pos - near_width))
        # COMPAT.
        elif cor_side == "near":
            deprecation_warning(
                "using side='near' is deprecated, use side=<a scalar> instead",
                do_print=True,
                func_name="composite_near_pos",
            )
            near_pos = self.cor_options["near_pos"]
            near_width = self.cor_options["near_width"]
            overlap_min = max(4, radio_width - 2 * self.ovs * (near_pos + near_width))
            overlap_max = min(2 * radio_width - 4, radio_width - 2 * self.ovs * (near_pos - near_width))
        # ---
        else:
            raise ValueError("Invalid option 'side=%s'" % self.cor_options["side"])

        if overlap_min > overlap_max:
            message = f""" There is no safe search range in find_cor once the margins corresponding to the high_pass filter are discarded.
            Try reducing the low_pass parameter in cor_options
            """
            raise ValueError(message)

        # cor_min and cor_max are only for information, not used in the algorithm, they correspond to the search window in the original projection grid (before oversampling)
        cor_min = detector_center + (radio_width - overlap_max) / (2 * self.ovs)
        cor_max = detector_center + (radio_width - overlap_min) / (2 * self.ovs)
        self.logger.info(f"looking around {cor_side} overlap from min {cor_min:2f} and max {cor_max:2f}")

        best_overlap = overlap_min
        best_error = np.inf

        blurred_radio1 = nd.gaussian_filter(abs(radio1), [0, self.high_pass])
        blurred_radio2 = nd.gaussian_filter(abs(radio2), [0, self.high_pass])

        for z in range(int(overlap_min), int(overlap_max) + 1):
            if z <= radio_width:
                my_z = z
                my_radio1 = radio1
                my_radio2 = radio2
                my_blurred_radio1 = blurred_radio1
                my_blurred_radio2 = blurred_radio2
            else:
                my_z = radio_width - (z - radio_width)
                my_radio1 = np.fliplr(radio1)
                my_radio2 = np.fliplr(radio2)
                my_blurred_radio1 = np.fliplr(blurred_radio1)
                my_blurred_radio2 = np.fliplr(blurred_radio2)

            common_left = np.fliplr(my_radio1[:, radio_width - my_z :])[:, : -math.ceil(self.ovs * self.high_pass * 2)]
            # adopt a 'safe' margin considering high_pass value (possibly float)
            common_right = my_radio2[:, radio_width - my_z : -math.ceil(self.ovs * self.high_pass * 2)]

            common_blurred_left = np.fliplr(my_blurred_radio1[:, radio_width - my_z :])[
                :, : -math.ceil(self.ovs * self.high_pass * 2)
            ]
            # adopt a 'safe' margin considering high_pass value (possibly float)
            common_blurred_right = my_blurred_radio2[:, radio_width - my_z : -math.ceil(self.ovs * self.high_pass * 2)]

            if common_right.size == 0:
                continue

            error = self.error_metric(common_right, common_left, common_blurred_right, common_blurred_left)

            min_error = min(best_error, error)

            if min_error == error:
                best_overlap = z
                best_error = min_error
            # self.logger.debug(
            #     "testing an overlap of %.2f pixels, actual best overlap is %.2f pixels over %d\r"
            #     % (z / self.ovs, best_overlap / self.ovs, ovsd_sx / self.ovs),
            # )

        offset = (radio_width - best_overlap) / self.ovs / 2
        cor_abs = (self.sx - 1) / 2 + offset

        return cor_abs

    def error_metric(self, common_right, common_left, common_blurred_right, common_blurred_left):
        if self.norm_order == 2:
            return self.error_metric_l2(common_right, common_left)
        elif self.norm_order == 1:
            return self.error_metric_l1(common_right, common_left, common_blurred_right, common_blurred_left)
        else:
            raise RuntimeError("this cannot happen")

    def error_metric_l2(self, common_right, common_left):
        common = common_right - common_left

        tmp = np.linalg.norm(common)
        norm_diff2 = tmp * tmp

        norm_right = np.linalg.norm(common_right)
        norm_left = np.linalg.norm(common_left)

        res = norm_diff2 / (norm_right * norm_left)

        return res

    def error_metric_l1(self, common_right, common_left, common_blurred_right, common_blurred_left):
        common = (common_right - common_left) / (common_blurred_right + common_blurred_left)

        res = abs(common).mean()

        return res


def oversample(radio, ovs_s):
    """oversampling an image in arbitrary directions.
    The first and last point of each axis will still remain as  extremal points of the new axis.
    """
    result = np.zeros([(radio.shape[0] - 1) * ovs_s[0] + 1, (radio.shape[1] - 1) * ovs_s[1] + 1], "f")

    # Pre-initialisation: The original data falls exactly on the following strided positions in the new data array.
    result[:: ovs_s[0], :: ovs_s[1]] = radio

    for k in range(ovs_s[0]):
        # interpolation coefficient for axis 0
        g = k / ovs_s[0]
        for i in range(ovs_s[1]):
            if i == 0 and k == 0:
                # this case subset was already exactly matched from before the present double loop,
                # in the pre-initialisation line.
                continue
            # interpolation coefficent for axis 1
            f = i / ovs_s[1]

            # stop just a bit before cause we are not extending beyond the limits.
            # If we are exacly on a vertical or horizontal original line, then no shift will be applied,
            # and we will exploit the equality f+(1-f)=g+(1-g)=1 adding twice the same contribution with
            # interpolation factors which become dummies pour le coup.
            stop0 = -ovs_s[0] if k else None
            stop1 = -ovs_s[1] if i else None

            # Once again, we exploit the  g+(1-g)=1 equality
            start0 = ovs_s[0] if k else 0
            start1 = ovs_s[1] if i else 0

            # and what is done below makes clear the corundum above.
            result[k :: ovs_s[0], i :: ovs_s[1]] = (1 - g) * (
                (1 - f) * result[0 : stop0 : ovs_s[0], 0 : stop1 : ovs_s[1]]
                + f * result[0 : stop0 : ovs_s[0], start1 :: ovs_s[1]]
            ) + g * (
                (1 - f) * result[start0 :: ovs_s[0], 0 : stop1 : ovs_s[1]]
                + f * result[start0 :: ovs_s[0], start1 :: ovs_s[1]]
            )
    return result


# alias
CompositeCOREstimator = CompositeCORFinder


# Some heavily inelegant things going on here
def get_default_kwargs(func):
    params = inspect.signature(func).parameters
    res = {}
    for param_name, param in params.items():
        if param.default != inspect._empty:
            res[param_name] = param.default
    return res


def update_func_kwargs(func, options):
    res_options = get_default_kwargs(func)
    for option_name, option_val in options.items():
        if option_name in res_options:
            res_options[option_name] = option_val
    return res_options


def get_class_name(class_object):
    return str(class_object).split(".")[-1].strip(">").strip("'").strip('"')


class DetectorTiltEstimator:
    """
    Helper class for detector tilt estimation.
    It automatically chooses the right radios and performs flat-field.
    """

    default_tilt_method = "1d-correlation"
    # Given a tilt angle "a", the maximum deviation caused by the tilt (in pixels) is
    #  N/2 * |sin(a)|  where N is the number of pixels
    # We ignore tilts causing less than 0.25 pixel deviation: N/2*|sin(a)| < tilt_threshold
    tilt_threshold = 0.25

    def __init__(self, dataset_info, do_flatfield=True, logger=None, autotilt_options=None):
        """
        Initialize a detector tilt estimator helper.

        Parameters
        ----------
        dataset_info: `dataset_info` object
            Data structure with the dataset information.
        do_flatfield: bool, optional
            Whether to perform flat field on radios.
        logger: `Logger` object, optional
            Logger object
        autotilt_options: dict, optional
            named arguments to pass to the detector tilt estimator class.
        """
        self._set_params(dataset_info, do_flatfield, logger, autotilt_options)
        self.radios, self.radios_indices = get_radio_pair(dataset_info, radio_angles=(0.0, np.pi), return_indices=True)
        self._init_flatfield()
        self._apply_flatfield()

    def _set_params(self, dataset_info, do_flatfield, logger, autotilt_options):
        self.dataset_info = dataset_info
        self.do_flatfield = bool(do_flatfield)
        self.logger = LoggerOrPrint(logger)
        self._get_autotilt_options(autotilt_options)

    def _init_flatfield(self):
        if not (self.do_flatfield):
            return
        self.flatfield = FlatField(
            self.radios.shape,
            flats=self.dataset_info.flats,
            darks=self.dataset_info.darks,
            radios_indices=self.radios_indices,
            interpolation="linear",
        )

    def _apply_flatfield(self):
        if not (self.do_flatfield):
            return
        self.flatfield.normalize_radios(self.radios)

    def _get_autotilt_options(self, autotilt_options):
        if autotilt_options is None:
            self.autotilt_options = None
            return
        try:
            autotilt_options = extract_parameters(autotilt_options)
        except Exception as exc:
            msg = "Could not extract parameters from autotilt_options: %s" % (str(exc))
            self.logger.fatal(msg)
            raise ValueError(msg)
        self.autotilt_options = autotilt_options
        if "threshold" in autotilt_options:
            self.tilt_threshold = autotilt_options.pop("threshold")

    def find_tilt(self, tilt_method=None):
        """
        Find the detector tilt.

        Parameters
        ----------
        tilt_method: str, optional
            Which tilt estimation method to use.
        """
        if tilt_method is None:
            tilt_method = self.default_tilt_method
        check_supported(tilt_method, set(tilt_methods.values()), "tilt estimation method")
        self.logger.info("Estimating detector tilt angle")
        autotilt_params = {
            "roi_yxhw": None,
            "median_filt_shape": None,
            "padding_mode": None,
            "peak_fit_radius": 1,
            "high_pass": None,
            "low_pass": None,
        }
        autotilt_params.update(self.autotilt_options or {})
        self.logger.debug("%s(%s)" % ("CameraTilt", str(autotilt_params)))

        tilt_calc = CameraTilt()
        tilt_cor_position, camera_tilt = tilt_calc.compute_angle(
            self.radios[0], np.fliplr(self.radios[1]), method=tilt_method, **autotilt_params
        )
        self.logger.info("Estimated detector tilt angle: %f degrees" % camera_tilt)
        # Ignore too small tilts
        max_deviation = np.max(self.dataset_info.radio_dims) * np.abs(np.sin(np.deg2rad(camera_tilt)))
        if self.dataset_info.is_halftomo:
            max_deviation *= 2
        if max_deviation < self.tilt_threshold:
            self.logger.info(
                "Estimated tilt angle (%.3f degrees) results in %.2f maximum pixels shift, which is below threshold (%.2f pixel). Ignoring the tilt, no correction will be done."
                % (camera_tilt, max_deviation, self.tilt_threshold)
            )
            camera_tilt = None
        return camera_tilt


# alias
TiltFinder = DetectorTiltEstimator


def estimate_translations(dataset_info, do_flatfield=True): ...


class TranslationsEstimator:

    _default_extra_options = {
        "window_size": 300,
    }

    def __init__(
        self,
        dataset_info,
        do_flatfield=True,
        rot_center=None,
        halftomo_side=None,
        angular_subsampling=10,
        deg_xy=2,
        deg_z=2,
        shifts_estimator="phase_cross_correlation",
        radios_filter=None,
        extra_options=None,
    ):
        self._configure_extra_options(extra_options)
        self.logger = LoggerOrPrint(dataset_info.logger)
        self.dataset_info = dataset_info
        self.angular_subsampling = angular_subsampling
        self.do_360 = self.dataset_info.is_360
        self.do_flatfield = do_flatfield
        self.radios_filter = radios_filter
        self.radios = None
        self._deg_xy = deg_xy
        self._deg_z = deg_z
        self._shifts_estimator = shifts_estimator
        self._shifts_estimator_kwargs = {}
        self._cor = rot_center
        self._configure_halftomo(halftomo_side)
        self._estimate_cor = self._cor is None
        self.sample_shifts_xy = None
        self.sample_shifts_z = None

    def _configure_extra_options(self, extra_options):
        self.extra_options = self._default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def _configure_halftomo(self, halftomo_side):
        if halftomo_side is False:
            # Force disable halftomo
            self.halftomo_side = False
            return
        self._start_x = None
        self._end_x = None
        if (halftomo_side is not None) and not (self.do_360):
            raise ValueError(
                "Expected 360° dataset for half-tomography, but this dataset does not look like a 360° dataset"
            )
        if halftomo_side is None:
            if self.dataset_info.is_halftomo:
                halftomo_side = "right"
            else:
                self.halftomo_side = False
                return
        self.halftomo_side = halftomo_side
        window_size = self.extra_options["window_size"]
        if self._cor is not None:
            # In this case we look for shifts around the CoR
            self._start_x = int(self._cor - window_size / 2)
            self._end_x = int(self._cor + window_size / 2)
        elif halftomo_side == "right":
            self._start_x = -window_size
            self._end_x = None
        elif halftomo_side == "left":
            self._start_x = 0
            self._end_x = window_size
        elif is_scalar(halftomo_side):
            # Expect approximate location of CoR, relative to left-most column
            self._start_x = int(halftomo_side - window_size / 2)
            self._end_x = int(halftomo_side + window_size / 2)
        else:
            raise ValueError(
                f"Expected 'halftomo_side' to be either 'left', 'right', or an integer (got {halftomo_side})"
            )
        self.logger.debug(f"[MotionEstimation] Half-tomo looking at [{self._start_x}:{self._end_x}]")
        # For half-tomo, skimage.registration.phase_cross_correlation might look a bit too far away
        if (
            self._shifts_estimator == "phase_cross_correlation"
            and self._shifts_estimator_kwargs.get("overlap_ratio", 0.3) >= 0.3
        ):
            self._shifts_estimator_kwargs.update({"overlap_ratio": 0.2})
        #

    def _load_data(self):
        self.logger.debug("[MotionEstimation] reading data")
        if self.do_360:
            """
            In this case we compare pair of opposite projections.
            If rotation angles are arbitrary, we should do something like
              for angle in dataset_info.rotation_angles:
                  img, angle_deg, idx = dataset_info.get_image_at_angle(
                      np.degrees(angle)+180, return_angle_and_index=True
                  )
            Most of the time (always ?), the dataset was acquired with a circular trajectory,
            so we can use angles:
                dataset_info.rotation_angles[::self.angular_subsampling]
            which amounts to reading one radio out of "angular_subsampling"
            """

            # TODO account for more general rotation angles. The following will only work for circular trajectory and ordered angles
            self._reader = self.dataset_info.get_reader(
                sub_region=(slice(None, None, self.angular_subsampling), slice(None), slice(None))
            )
            self.radios = self._reader.load_data()
            self.angles = self.dataset_info.rotation_angles[:: self.angular_subsampling]
            self._radios_idx = self._reader.get_frames_indices()
            self.logger.debug("[MotionEstimation] This is a 360° scan, will use pairs of opposite projections")
        else:
            """
            In this case we use the "return projections", i.e special projections acquired at several angles
            (eg. [180, 90, 0]) before ending the scan
            """
            return_projs, return_angles_deg, return_idx = self.dataset_info.get_alignment_projections()
            self._angles_return = np.radians(return_angles_deg)
            self._radios_return = return_projs
            self._radios_idx_return = return_idx

            projs = []
            angles_rad = []
            projs_idx = []
            for angle_deg in return_angles_deg:
                proj, rot_angle_deg, proj_idx = self.dataset_info.get_image_at_angle(
                    angle_deg, image_type="projection", return_angle_and_index=True
                )
                projs.append(proj)
                angles_rad.append(np.radians(rot_angle_deg))
                projs_idx.append(proj_idx)
            self._radios_outwards = np.array(projs)
            self._angles_outward = np.array(angles_rad)
            self._radios_idx_outwards = np.array(projs_idx)
            self.logger.debug("[MotionEstimation] This is a 180° scan, will use 'return projections'")

    def _apply_flatfield(self):
        if not (self.do_flatfield):
            return
        self.logger.debug("[MotionEstimation] flatfield")
        if self.do_360:
            self._flatfield = FlatField(
                self.radios.shape,
                flats=self.dataset_info.flats,
                darks=self.dataset_info.darks,
                radios_indices=self._radios_idx,
            )
            self._flatfield.normalize_radios(self.radios)
        else:
            # 180 + return projs
            self._flatfield_outwards = FlatField(
                self._radios_outwards.shape,
                flats=self.dataset_info.flats,
                darks=self.dataset_info.darks,
                radios_indices=self._radios_idx_outwards,
            )
            self._flatfield_outwards.normalize_radios(self._radios_outwards)
            self._flatfield_return = FlatField(
                self._radios_return.shape,
                flats=self.dataset_info.flats,
                darks=self.dataset_info.darks,
                radios_indices=self._radios_idx_return,
            )
            self._flatfield_outwards.normalize_radios(self._radios_return)

    def estimate_motion(self):
        self._load_data()
        self._apply_flatfield()
        if self.radios_filter is not None:
            self.logger.debug("[MotionEstimation] applying radios filter")
            self.radios_filter(self.radios)

        n_projs_tot = self.dataset_info.n_angles
        if self.do_360:
            n_a = self.radios.shape[0]
            # See notes above - this works only for circular trajectory / ordered angles
            projs_stack1 = self.radios[: n_a // 2]
            projs_stack2 = self.radios[n_a // 2 :]
            angles1 = self.angles[: n_a // 2]
            angles2 = self.angles[n_a // 2 :]
            indices1 = (self._radios_idx - self._radios_idx[0])[: n_a // 2]
            indices2 = (self._radios_idx - self._radios_idx[0])[n_a // 2 :]
        else:
            projs_stack1 = self._radios_outwards
            projs_stack2 = self._radios_return
            angles1 = self._angles_outward
            angles2 = self._angles_return
            indices1 = self._radios_idx_outwards - self._radios_idx_outwards.min()
            indices2 = self._radios_idx_return - self._radios_idx_outwards.min()

        if self._start_x is not None:
            # Compute Motion Estimation on subset of images (eg. for half-tomo)
            projs_stack1 = projs_stack1[..., self._start_x : self._end_x]
            projs_stack2 = projs_stack2[..., self._start_x : self._end_x]

        self.motion_estimator = MotionEstimation(
            projs_stack1,
            projs_stack2,
            angles1,
            angles2,
            indices1,
            indices2,
            n_projs_tot,
            shifts_estimator=self._shifts_estimator,
            shifts_estimator_kwargs=self._shifts_estimator_kwargs,
        )

        self.logger.debug("[MotionEstimation] estimating shifts")

        estimated_shifts_v = self.motion_estimator.estimate_vertical_motion(degree=self._deg_z)
        estimated_shifts_h, cor = self.motion_estimator.estimate_horizontal_motion(degree=self._deg_xy, cor=self._cor)
        if self._start_x is not None:
            cor += (self._start_x % self.radios.shape[-1]) + (projs_stack1.shape[-1] - 1) / 2.0

        self.sample_shifts_xy = estimated_shifts_h
        self.sample_shifts_z = estimated_shifts_v
        if self._cor is None:
            self.logger.info(
                "[MotionEstimation] Estimated center of rotation (relative to left-most pixel): %.2f" % cor
            )
        return estimated_shifts_h, estimated_shifts_v, cor

    def generate_translations_movements_file(self, filename, fmt="%.3f", only=None):
        if self.sample_shifts_xy is None:
            raise RuntimeError("Need to run estimate_motion() first")

        angles = self.dataset_info.rotation_angles
        cor = self._cor or 0
        txy_est_all_angles = self.motion_estimator.apply_fit_horiz(angles=angles)
        tz_est_all_angles = self.motion_estimator.apply_fit_vertic(angles=angles)
        estimated_shifts_vu_all_angles = self.motion_estimator.convert_sample_motion_to_detector_shifts(
            txy_est_all_angles, tz_est_all_angles, angles, cor=cor
        )
        estimated_shifts_vu_all_angles[:, 1] -= cor
        correct_shifts_uv = -estimated_shifts_vu_all_angles[:, ::-1]

        if only is not None:
            if only == "horizontal":
                correct_shifts_uv[:, 1] = 0
            elif only == "vertical":
                correct_shifts_uv[:, 0] = 0
            else:
                raise ValueError("Expected 'only' to be either None, 'horizontal' or 'vertical'")

        header = f"Generated by nabu {nabu_version} : {str(self)}"  # noqa: RUF010
        np.savetxt(filename, correct_shifts_uv, fmt=fmt, header=header)

    def __str__(self):
        ret = f"{self.__class__.__name__}(do_flatfield={self.do_flatfield}, rot_center={self._cor}, angular_subsampling={self.angular_subsampling})"
        if self.sample_shifts_xy is not None:
            ret += f", shifts_estimator={self.motion_estimator.shifts_estimator}"
        return ret



import os
from nabu import version as nabu_version

# This file is used by the "gen-files" mkdocs plugin
# It writes _data/variables.yml at build time
# which is then used by the "markdownextradata" plugin

template = """
nabu:
  version: {nabu_version}
"""


if not (os.path.exists("_data")):
    os.mkdir("_data")
with open(os.path.join("_data", "variables.yml"), "w") as f:
    content = template.format(nabu_version=nabu_version)
    f.write(content)
